import copy
import time
import warnings
import re
import datetime
import json
import pathlib

from .login import Login
from .. import sxs_id, Metadata
from ..utilities import sxs_identifier_regex, SimpleVersion


# To do:
# - Finish up the code to create a new version from an existing one


def mtime(f):
    """Look for git or filesystem modification time

    If the git call fails for any reason, we fall back to the filesystem's `mtime`.

    The input `f` is expected to be a path to the file.  The return value is a DateTime object

    """
    from subprocess import check_output
    try:
        timestamp = check_output(f"""git log -1 --format="%ad" --date=unix -- {f}""", shell=True).decode().strip()
    except:
        timestamp = f.stat().st_mtime
    finally:
        mt = datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)
    return mt


def filter_files(files, utime):
    return [f for f in files if mtime(f) > utime]


def upload_simannex_dir(login, directory, date=None):
    """Update/upload data from the SimAnnex to CaltechDATA

    Parameters
    ----------
    login : caltechdata.Login
        This is used to communicate with CaltechDATA.
    directory : str or pathlib.Path
        This must contain either "Public" or "Private", followed by the specific
        directory in SimAnnex to upload.  It can also contain higher path elements
        — for example, you can give the absolute path.
    date : str, optional
        If given, this is used as the `publicationDate` field.  It must be
        formatted as "%Y-%m-%d".  The default is just the current UTC time.

    """
    raise Exception("This interface is for the old TIND API; use zenodo for the new API")
    from caltechdata_api import customize_schema, decustomize_schema

    allowed_files = [
        "metadata.json",
        "Strain_Outer.h5",
        "Strain_N2.h5",
        "Strain_N3.h5",
        "Strain_N4.h5",
        "Horizons.h5",
        "Matter.h5",
    ]

    # Check that `directory` is a valid SXS dataset
    # - Name must contain either "Public" or "Private" followed by at least one subdirectory
    # - Must contain common-metadata.txt (though this will not be uploaded)
    # - Must contain at least one Lev* subdirectory
    # - Each Lev* subdirectory must contain these files (which will be uploaded)
    #   - metadata.json
    #   - Strain_Outer.h5
    #   - Strain_N2.h5
    #   - Strain_N3.h5
    #   - Strain_N4.h5
    #   - Horizons.h5 and/or Matter.h5
    directory = pathlib.Path(directory).expanduser().resolve()
    if "/Public/" not in str(directory) and "/Private/" not in str(directory):
        raise ValueError(f"""Either "Public" or "Private" must be in the input directory "{directory}".""")
    if not (directory / "common-metadata.txt").exists():
        raise ValueError(f"Missing common-metadata.txt in {directory}")
    if not directory.glob("Lev*"):
        raise ValueError(f"Missing Lev* in {directory}")
    for lev in directory.glob("Lev*"):
        warnings.warn("Temporarily skipping file checks")
        # for f in ["Strain_Outer.h5", "Strain_N2.h5", "Strain_N3.h5", "Strain_N4.h5"]:
        #     if not (lev / f).exists():
        #         raise ValueError(f"""Missing file "{f}" in "{lev}".""")
        # if not (lev / "metadata.json").exists():
        #     mtxt = lev / "metadata.txt"
        #     if not mtxt.exists():
        #         raise ValueError(f"""Missing both "metadata.txt" and "metadata.json" in "{lev}".""")
        #     else:
        #         m = Metadata.from_txt_file(str(mtxt), cache_json=False).add_extras().reorder_keys()
        #         m.to_json_file(str(mtxt.with_suffix(".json")))
        # if not (lev / "Horizons.h5").exists() and not (lev / "Matter.h5").exists():
        #     raise ValueError(f"""Missing both "Horizons.h5" and "Matter.h5" in "{lev}".""")

    # Figure out the "original name" of this simulation — something like "q4_7d/022" or "HighSpinMethods/BBH_SKS_d15.4_q1_..."
    root_dir = [par for par in directory.parents if par.match("Public") or par.match("Private")][0]
    original_name = str(directory.relative_to(root_dir))

    # Get the SXS ID from common-metadata.txt
    sxs_system_re = re.compile(sxs_identifier_regex)
    sxs_system_type = None
    sxs_system_number = None
    with (directory / "common-metadata.txt").open("r") as f:
        for line in f.readlines():
            line = line.strip()
            if "alternative-names" in line:
                m = sxs_system_re.search(line)
                if m:
                    sxs_system_type = m["simulation_type"]
                    sxs_system_number = m["sxs_number"]
                    break
    if not sxs_system_type or not sxs_system_number:
        raise ValueError(f"No SXS identifier found in {directory / 'common-metadata.txt'}")
    sxs_system = f"SXS:{sxs_system_type}:{sxs_system_number}"
    spec_url = "https://www.black-holes.org/code/SpEC.html"
    if sxs_system_type == "BBH":
        title = f"Binary black-hole simulation {sxs_system}"
        description = f"""Simulation of a black-hole binary system evolved by the <a href="{spec_url}">SpEC code</a>."""
    elif sxs_system_type == "BHNS":
        title = f"Black-hole neutron-star binary simulation {sxs_system}"
        description = f"""Simulation of a black-hole neutron-star binary system evolved by the <a href="{spec_url}">SpEC code</a>."""
    elif sxs_system_type == "NSNS":
        title = f"Binary neutron-star simulation {sxs_system}"
        description = f"""Simulation of a neutron-star binary system evolved by the <a href="{spec_url}">SpEC code</a>."""
    else:
        raise ValueError(
            f"""Did not recognize SXS system type "{sxs_system_type}" in directory "{directory}"; should be BBH, BHNS, or NSNS."""
        )
    print(f"Beginning work on {sxs_system}")

    # Find our list of files
    files = sorted(
        [f for f in directory.glob("Lev*/*") if f.name in allowed_files],
        key=lambda f: f"{f.parent}/{allowed_files.index(f.name)}"
    )

    # Search for an existing record with this name
    existing = login.search(f'"{sxs_system}"')
    exists = existing.get("hits", {}).get("total", 0) > 0

    if exists:
        # Find the latest version
        latest = max(
            existing["hits"]["hits"],
            key=lambda h: SimpleVersion(h["metadata"]["version"])
        )
        latest_link = latest["links"]["self"].replace("/api/", "/")
        latest_doi = latest["metadata"]["doi"]

        # Check to see if any files have changed since the latest version
        utime = datetime.datetime.fromisoformat(latest["updated"])
        files_to_upload = filter_files(files, utime)
        if not files:
            print(f"No changes to {latest_link}.")
            return latest_link

        # Update version, relatedIdentifiers, and updated date
        metadata = latest["metadata"]
        old_version = SimpleVersion(metadata["version"])
        metadata["version"] = str(old_version.increment())
        metadata["relatedIdentifiers"] =[
            {
                "relatedIdentifierRelation": "IsNewVersionOf",
                "relatedIdentifier": latest_doi,
                "relatedIdentifierScheme": "DOI"
            }
        ]
        metadata["relevantDates"] = [
            {
                "relevantDateType": "Updated",
                "relevantDateValue": datetime.datetime.utcnow().strftime("%Y-%m-%d"),
            }
        ]

        # Upload files
        raise NotImplementedError(
            "Adding new files to an existing record is not yet supported.  We will need some logic like this:\n" +
            "  https://github.com/caltechlibrary/caltechdata_api/blob/840a359a017d1257ac344b6af80038f591fd8a97/caltechdata_api/caltechdata_edit.py#L189-L233\n" +
            "More specifically, if we want to create a new version, we will probably need to essentially copy\n" +
            "the existing record, publish it, and then follow steps like the above to upload new files and\n" +
            "delete the old ones.  It is not clear to me if the deletion would destroy files being used by the\n" +
            "older version, or just delete that entry in the new record."
        )
        existing_names = {
            f["electronic_name"][0]: f["uniform_resource_identifier"].split("/")[-2]
            for f in metadata["electronic_location_and_access"]
        }
        for f in files_to_upload:
            fileinfo = login.send_s3(f, str(f.relative_to(directory)), verbose=True)
            name = fileinfo["filename"]
            if name in existing_names:
                metadata["files"][existing_names.index(name)] = fileinfo
            else:
                metadata["files"].append(fileinfo)

    else:
        # Set up the metadata.  Note that CaltechDATA internally uses an undocumented
        # customized schema similar to — but different from — the standard Invenio
        # schema.  Fortunately, `caltechdata_api` includes a function to convert from
        # DataCite 4.0 or 4.3 to their schema.  So, we set ours up as DataCite 4.0 and
        # convert it: <https://schema.datacite.org/meta/kernel-4.0/>

        # Generate a little table of info to append to the description
        highest_lev = sorted(set(f.relative_to(directory).parent for f in files))[-1]
        sim_metadata = Metadata.from_file(directory / highest_lev / "metadata")
        object_types = sim_metadata.get("object_types", "")
        mass_ratio = sim_metadata.get("reference_mass_ratio", "Unknown")
        if "reference_dimensionless_spin1" in sim_metadata:
            chi1 = sim_metadata.reference_dimensionless_spin1
            chi1 = f"[{chi1[0]:0.4f}, {chi1[1]:0.4f}, {chi1[2]:0.4f}]"
        else:
            chi1 = "N/A"
        if "reference_dimensionless_spin2" in sim_metadata:
            chi2 = sim_metadata.reference_dimensionless_spin2
            chi2 = f"[{chi2[0]:0.4f}, {chi2[1]:0.4f}, {chi2[2]:0.4f}]"
        else:
            chi2 = "N/A"
        n_orbits = sim_metadata.get("number_of_orbits", "Unknown")
        reference_eccentricity = sim_metadata.get("reference_eccentricity", "Unknown")
        description = f"""{description}
        <br/>
        <table>
          <tbody>
            <tr><td style="padding:0 15px;">Mass ratio</td><td>{mass_ratio:.4f}</td></tr>
            <tr><td style="padding:0 15px;">Spin 1</td><td>{chi1}</td></tr>
            <tr><td style="padding:0 15px;">Spin 2</td><td>{chi2}</td></tr>
            <tr><td style="padding:0 15px;">Number of orbits</td><td>{n_orbits:.4f}</td></tr>
            <tr><td style="padding:0 15px;">Eccentricity</td><td>{reference_eccentricity}</td></tr>
          </tbody>
        </table>
        <br/>
        Originally named "{original_name}"
        """
        files_to_upload = files

        metadata = {
            "resourceType": {
                "resourceTypeGeneral": "Dataset"
            },
            "titles": [
                {"title": title},
            ],
            "version": "2.0",
            "creators": [{"creatorName": "SXS Collaboration"}],
            "descriptions": [
                {
                    "descriptionType": "Abstract",
                    "description": description,
                }
            ],
            "subjects": [
                {"subject": "Astronomy"},
                {"subject": "Astrophysics"},
                {"subject": "Gravitational Waves"},
                {"subject": "Numerical Relativity"},
            ],
            "dates": [
                {
                    "date": datetime.datetime.utcnow().strftime("%Y-%m-%d"),
                    "dateType": "Updated"
                },
            ],
            "rightsList": [
                {
                    "rights": "Creative Commons Attribution 4.0 International License",
                    "rightsURI": "https://creativecommons.org/licenses/by/4.0/",
                }
            ],
        }
        if "bh" in object_types.lower():
            metadata["subjects"].append({"subject": "Black Holes"})
        if "ns" in object_types.lower():
            metadata["subjects"].append({"subject": "Neutron Stars"})
        metadata = customize_schema(metadata, schema="40")
        metadata["publicationDate"] = date or datetime.datetime.utcnow().strftime("%Y-%m-%d")
        metadata["files"] = [login.send_s3(f, str(f.relative_to(directory)), verbose=True) for f in files]

    metadata["titles"] = [
        {"title": title},
        {"title": f"{original_name}", "titleType": "AlternativeTitle"},
    ]

    # if doi is None:
    #     # We want tind to generate the identifier
    #     metadata["final_actions"] = [
    #         {
    #             "type": "create_doi",
    #             "parameters": {"type": "records", "field": "doi"},
    #         }
    #     ]
    # else:
    metadata["doi"] = f"""{login.doi_prefix}/{sxs_system}v{metadata["version"]}"""

    # Now tell CaltechDATA about it
    recordurl = f"{login.base_url}submit/api/create/"
    response = login.session.post(recordurl, data=json.dumps({"record": metadata}))
    if response.status_code != 200:
        print(f"An error occurred when trying to create a new record for '{directory}'.")
        try:
            print(response.text)
        except:
            pass
        try:
            print(response.json())
        except:
            pass
        response.raise_for_status()
        raise RuntimeError()  # Will only happen if the response was not strictly an error
    else:
        print(f"""  Publishing as "{title}".""")
        print(f"""  {response.text}""")

    url = response.text[response.text.find(login.base_url):].rstrip().rstrip(".")
    pid = url.rsplit("/", 1)[1]

    # Create the DOI
    print("  Creating/updating DOIs")
    api_url = f"{login.base_url}api/record/{pid}"
    for retry in range(20):
        r = login.session.get(api_url)
        if r.status_code == 200:
            break
        time.sleep(1.1)  # Let things settle down at CaltechDATA, so we can get the metadata
    if r.status_code != 200:
        print(f"""An error occurred when trying to access "{api_url}".""")
        try:
            print(r.json())
        except:
            pass
        r.raise_for_status()
        raise RuntimeError()  # Will only happen if the response was not strictly an error
    output_metadata = r.json()
    doi_metadata = decustomize_schema(output_metadata, schema="43")
    if not exists:
        doi = f"""{login.doi_prefix}/{sxs_system}"""
        login.datacite.public_doi(doi_metadata, url, doi)
    doi = f"""{login.doi_prefix}/{sxs_system}v{metadata["version"]}"""
    login.datacite.public_doi(doi_metadata, url, doi)

    return url



"""Containers for various types of waveforms

Currently, the most interesting object in this submodule is
[`WaveformModes`](../sxs.waveforms.waveform_modes/).

"""

from .waveform_mixin import WaveformMixin
from .waveform_modes import WaveformModes, WaveformModesDict
# from .waveform_grid import WaveformGrid
# from .waveform_signal import WaveformSignal

from .format_handlers import (
    nrar,
    rotating_paired_diff_multishuffle_bzip2,
    rotating_paired_xor_multishuffle_bzip2,
    spectre_cce_v1,
)
from .format_handlers.lvc import to_lvc_conventions
from . import memory, transformations, alignment

# Map a format string to the module that should be used to load a waveform with that format
formats = {
    None: nrar,
    "": nrar,
    "nrar": nrar,
    "rotating_paired_xor_multishuffle_bzip2": rotating_paired_xor_multishuffle_bzip2,
    "rpxm": rotating_paired_xor_multishuffle_bzip2,
    "rpxmb": rotating_paired_xor_multishuffle_bzip2,
    "rotating_paired_diff_multishuffle_bzip2": rotating_paired_diff_multishuffle_bzip2,
    "rpdmb": rotating_paired_diff_multishuffle_bzip2,
    "SpECTRE_CCE_v1": spectre_cce_v1,
}



import numpy as np
import sxs
from sxs import julia

def test_PNWaveform():
    M1 = 0.3
    M2 = 0.7
    chi1 = [0.1, 0.2, 0.3]
    chi2 = [-0.2, 0.1, -0.3]
    Omega_i = 0.01
    PNOrder = 4.0
    waveform_pn_order = 4.0
    saves_per_orbit = 32
    w1 = sxs.julia.PNWaveform(
        M1, M2, chi1, chi2, Omega_i,
        PNOrder=PNOrder, waveform_pn_order=waveform_pn_order,
        saves_per_orbit=saves_per_orbit
    )

    Approximant = "TaylorT1"
    delta = sxs.julia.PostNewtonian.mass_difference_ratio(M1, M2)
    chi1_i = chi1
    chi2_i = chi2
    Omega_orb_i = Omega_i
    inertial = True
    w2 = sxs.julia.GWFrames.PNWaveform(Approximant, delta, chi1_i, chi2_i, Omega_orb_i, inertial=inertial)

    for w in [w1, w2]:
        n_times = w.n_times
        assert w.t[0] == 0.0
        assert w.data.shape == (n_times, 77)
        assert w.data_type == "h"
        assert w.ell_min == 2
        assert w.ell_max == 8
        assert w.spin_weight == -2
        assert w.frame.shape == (1, 4)
        assert w.frame_type == "inertial"
        assert w.M1.shape == (n_times,)
        assert w.M2.shape == (n_times,)
        assert w.chi1.shape == (n_times, 3)
        assert w.chi2.shape == (n_times, 3)
        assert w.coorbital_frame.shape == (n_times, 4)
        assert w.v.shape == (n_times,)
        assert w.orbital_phase.shape == (n_times,)

    assert np.allclose(w1.t[:-10], w2.t[:-10])
    assert np.allclose(w1.data[:-10, :], w2.data[:-10, :], rtol=1e-3, atol=1e-5)
    assert np.allclose(w1.frame, w2.frame)
    assert np.allclose(w1.M1[:-10], w2.M1[:-10])
    assert np.allclose(w1.M2[:-10], w2.M2[:-10])
    assert np.allclose(w1.chi1[:-10, :], w2.chi1[:-10, :], rtol=1e-3)
    assert np.allclose(w1.chi2[:-10, :], w2.chi2[:-10, :], rtol=1e-3)
    assert np.allclose(w1.coorbital_frame[:-10, :], w2.coorbital_frame[:-10, :])

    w1_corot = w1.copy().to_corotating_frame()
    w2_corot = w2.copy().to_corotating_frame()
    for (w, w_corot) in [(w1, w1_corot), (w2, w2_corot)]:
        assert np.array_equal(w.t, w_corot.t)
        assert w.data.shape == w_corot.data.shape
        assert w.coorbital_frame.shape == w_corot.frame.shape
        assert w_corot.frame_type == "corotating"



"""Class encapsulating interactions with CaltechDATA

This class handles requests through the CaltechDATA web API.  In particular, it manages authorization,
and retries any failed requests automatically.

"""

import requests

class CaltechDATAAuth(requests.auth.AuthBase):
    def __init__(self, base_url, access_token):
        self.base_url = base_url
        self.access_token = access_token
        #super(CaltechDATAAuth, self).__init__()

    def __call__(self, request):
        if request.url.startswith(self.base_url):
            request.headers["Authorization"] = f"Bearer {self.access_token}"
        return request

    def __eq__(self, other):
        return all(
            [
                self.base_url == getattr(other, "base_url", None),
                self.access_token == getattr(other, "access_token", None),
            ]
        )

    def __ne__(self, other):
        return not self == other


class Login(object):
    def __init__(
        self,
        url="https://cd-sandbox.tind.io/",
        access_token=None,
        access_token_path=None,
        total_retry_count=50,
        backoff_factor=0.1,
        backoff_max=20.0,
        session=None,
    ):
        """Initialize a Login object for interacting with CaltechDATA

        This object encapsulates the credentials needed to interact with the CaltechDATA API, and
        exposes a Session object that can be used to make requests which automatically include the
        credentials.  It can be used for generic requests, but note that other objects in this
        module make certain tasks easier -- such as creating or modifying a "deposit", which is
        CaltechDATA's name for a new upload.  The Deposit object should be created from this object.

        These actions require a CaltechDATA API personal access token.  These are obtained from the
        website -- either
            https://data.caltech.edu/account/settings/applications/tokens/new/
        or
            https://cd-sandbox.tind.io/account/settings/applications/tokens/new/
        Note that these two options use separate login systems and separate access tokens.  The
        access token may either be passed as a string to this function (though that means it will
        probably be found in some script file somewhere, which is probably not a good idea for
        security), or can be read from a file.  By default the file from which the token is read is
        '~/.credentials/caltechdata/access_token' or the same name with '_sandbox' appended.  Thus,
        it is probably easiest to simply place your access tokens in those files, so that no
        arguments need to be passed to this function.  As a basic security measure, please ensure
        that those files are not readable by anyone but the user.

        Parameters
        ----------
        url : str [default: "https://cd-sandbox.tind.io/"]
            The base URL of the archive.  Note that the default URL is the "sandbox" version for
            CaltechDATA, which is just for testing purposes, and will likely be cleaned out
            regularly.  To upload to the archival site, pass its URL: "https://data.caltech.edu/".

        access_token: string or None [default: None]
            If present, this is used as the CaltechDATA API access token.

        access_token_path: string or None [default: None]
            If `access_token` is not given, this file is read and the first line is used as the
            access token.  If this argument is None, it defaults to either
            '~/.credentials/caltechdata/access_token' for the regular website or
            '~/.credentials/caltechdata/access_token_sandbox' for the sandbox website.

        total_retry_count: int [default: 50]
            Total number of times to retry requests that fail for retry-able reasons.

        backoff_factor: float [default: 0.1]
            A delay factor to apply between requests after the second try (most errors are resolved
            immediately by a second try without a delay).  After a certain number of total retries,
            the request Session will sleep for:

                {backoff factor} * (2 ^ ({number of total retries} - 1))

            seconds before trying again. For example, if the `backoff_factor` is 0.1, then the
            session will sleep for [0.0s, 0.2s, 0.4s, 0.8s, ...] between retries.  It will never be
            longer than `backoff_max`.

        backoff_max: float [default: 20.0]
            Longest time (in seconds) to wait between retries.

        session: requests.Session or None [default: None]
            This is the object that handles all of the requests made to the API.  If `None`, a
            Session is created for you, and sensible default headers (including the access token)
            are created.  If you need to adjust some of the Session parameters like proxies or SSL
            verification, you can simply create your own and pass it in here.  Note that any `auth`
            property on the passed object will be replaced by one that adds this to the header of
            each request to the chosen CaltechDATA domain:
                {"Authorization": "Bearer <YourAccessTokenHere>"}

        """
        import os
        import requests
        from requests.adapters import HTTPAdapter
        from urllib3.util.retry import Retry
        from datacite import DataCiteRESTClient

        self.base_url = url

        # The `session` object will handle all requests we make.
        self.session = session or requests.Session()

        # Set the CaltechDATA API access token
        if "sandbox" in url:
            access_token_path = "~/.credentials/caltechdata/access_token_sandbox"
            doi_auth_path = "~/.credentials/caltechdata/doi_auth_sandbox"
            self.doi_prefix = "10.80269"
            test_mode = True
        else:
            access_token_path = "~/.credentials/caltechdata/access_token"
            doi_auth_path = "~/.credentials/caltechdata/doi_auth"
            self.doi_prefix = "10.22002"
            test_mode = False
        with open(os.path.expanduser(doi_auth_path), "r") as f:
            user, password = f.readline().strip().split(":", 1)
        self.datacite = DataCiteRESTClient(
            username=user,
            password=password,
            prefix=self.doi_prefix,
            test_mode=test_mode,
        )
        path = os.path.expanduser(access_token_path)
        try:
            with open(path, "r") as f:
                self.access_token = f.readline().strip()
        except IOError:
            print("Unable to find the CaltechDATA access token needed to change a record.")
            print(f"Failed to open file '{path}' for reading.")
            raise
        if not self.access_token:
            print(f"The file '{path}' did not contain any text on the first line.")
            print("This is should be a CaltechDATA access token, which is needed to change a record.")
            raise ValueError("Deposit requires a CaltechDATA access token")

        # Ensure that this session sends the Authorization header with every request to the base_url
        self.session.auth = CaltechDATAAuth(self.base_url, self.access_token)

        # Note that some requests require different choices for 'Accept' and 'Content-Type'; these
        # are altered in the corresponding methods below.
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.session.headers.update(default_headers)

        ## Retry automatically on certain types of errors
        Retry.BACKOFF_MAX = backoff_max  # Must be set on the class, not the instance
        retry = Retry(
            total=total_retry_count,
            backoff_factor=backoff_factor,
            status_forcelist=[
                500,
                502,
                503,
                504,
            ],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount(self.base_url, adapter)

        # Test to see if we can use the given access token
        url = "{0}api/records".format(self.base_url)
        r = self.session.get(url)
        if r.status_code != 200:
            if r.status_code == 401:
                print(
                    f"The given CaltechDATA access token was not accepted by {self.base_url}.  Please ensure that it is still valid."
                )
                print(
                    "Also note that the standard site and the sandbox site use separate logins and separate access tokens."
                )
            else:
                print(f"An unknown error occurred when trying to access {self.base_url}.")
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an HTTP error

    def send_s3(self, path, name=None, verbose=False):
        import sys
        import pathlib
        #import tqdm

        if name is None:
            name = str(path)
        path = pathlib.Path(path).expanduser().resolve()
        size = path.stat().st_size

        if verbose:
            print(f"  Uploading {name} ({size:_} B) ", end="", flush=True)

        s3url = f"{self.base_url}tindfiles/sign_s3/"  # Note trailing slash
        chkurl = f"{self.base_url}tindfiles/md5_s3"

        r = self.session.get(s3url)
        if r.status_code != 200:
            if r.status_code == 401:
                print(
                    f"The given CaltechDATA access token was not accepted by {self.base_url}.  Please ensure that it is still valid."
                )
                print(
                    "Also note that the standard site and the sandbox site use separate logins and separate access tokens."
                )
                print(f"Used headers {r.request.headers}")
            else:
                print(f"An unknown error occurred when trying to access {self.base_url}.")
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an HTTP error
        s3 = r.json()
        data = s3["data"]
        bucket = s3["bucket"]

        key = data["fields"]["key"]
        policy = data["fields"]["policy"]
        aid = data["fields"]["AWSAccessKeyId"]
        signature = data["fields"]["signature"]
        url = data["url"]

        s3headers = {
            "Host": f"{bucket}.s3.amazonaws.com",
            "Date": "date",
            "x-amz-acl": "public-read",
            "Access-Control-Allow-Origin": "*",
        }

        with path.open("rb") as f:
            #with tqdm.tqdm.wrapattr(f, "read", total=size, desc="    ") as fw:  # To monitor upload progress
            form = (
                ("key", key),
                ("acl", "public-read"),
                ("AWSAccessKeyID", aid),
                ("policy", policy),
                ("signature", signature),
                ("file", f),
            )
            response = requests.session().post(url, files=form, headers=s3headers)
        if response.status_code != 204:
            if response.status_code == 400:
                print(f"Bad request: Probably caused by incorrectly formed input")
                print(f"Used headers {response.request.headers}")
            else:
                print(f"An unknown error occurred when trying to access {self.base_url}.")
            try:
                print(response.json())
            except:
                pass
            try:
                print(response.text)
            except:
                pass
            response.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an HTTP error

        response = self.session.get(f"{chkurl}/{bucket}/{key}/")
        md5 = response.json()["md5"]

        fileinfo = {"url": key, "filename": name, "md5": md5, "size": size}

        if verbose:
            print(f"✓")

        return fileinfo

    def download(self, url, path):
        """Download large file efficiently

        Parameters
        ----------
        url: string
            The URL to download from.  Redirects are followed.
        path: string
            Relative or absolute path to the file in which the download will be stored.  If this is
            an existing directory or ends in a path separator, the "path" component of the URL will
            be used as the file name, and the full directory path will be created.

        """
        from shutil import copyfileobj
        from os import makedirs
        from os.path import split, exists, join, isdir
        from functools import partial
        from urllib.parse import urlparse

        url_path = urlparse(url).path
        if isdir(path):
            path = join(path, url_path[1:])
            directory, filename = split(path)
            if not exists(directory):
                makedirs(directory)
            local_filename = join(directory, filename)
        else:
            directory, filename = split(path)
            if not exists(directory):
                makedirs(directory)
            if not filename:
                filename = url_path
            local_filename = join(directory, filename)
        r = self.session.get(url, stream=True, allow_redirects=True)
        if r.status_code != 200:
            print("An error occurred when trying to access <{0}>.".format(url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        r.raw.read = partial(r.raw.read, decode_content=True)
        # r.raw.decode_content = True
        with open(local_filename, "wb") as f:
            copyfileobj(r.raw, f)
        return local_filename

    @property
    def new_deposit(self):
        """Create a new Deposit object using this login"""
        return self.deposit()

    def deposit(self, deposition_id=None, ignore_deletion=False):
        """Retrieve a deposit created with this login"""
        from .deposit import Deposit

        return Deposit(self, deposition_id, ignore_deletion)

    def search(self, q=None, sort=None, size=1000, page=1, allversions=False, max_pages=10):
        """Return list of dictionaries describing each deposit created with this login

        It is possible to filter the results use the optional parameters.  Note that the web interface
        can sometimes be used to find search parameters by looking in the `search-hidden-params`
        parameter of the `invenio-search` tag.

        Example queries
        ---------------
        'title:"SXS:BBH:0003"'  # Finds titles with given string; use quotes for robustness
        'communities:sxs'  # Records in the 'sxs' CaltechDATA community
        'provisional_communities:sxs'  # Records awaiting approval by the community curator
        'owners: 38418'  # Find records by id number of owner

        Optional parameters
        -------------------
        q: string [optional]
            Search query, using Elasticsearch query string syntax.  See
            https://help.zenodo.org/guides/search/ for details.
        sort: string [optional]
            Sort order ('bestmatch' or 'mostrecent').  Prefix with minus to change from ascending to
            descending (e.g., '-mostrecent').
        size: int [optional, defaults to 1000]
            Number of results to return per page.  Note that CaltechDATA (as of this writing) seems to
            place a hard limit of 9999 responses.  Anything more will result in an error.  Use
            multiple pages to get more results.
        page: int [optional, defaults to 1]
            Page number for pagination
        allversions: bool [optional, defaults to False]
            If True return all records, including older versions of published records.
        max_pages: int [optional, defaults to 10]
            If the query returns a number of records equal to `size`, it is evidently incomplete.
            This function will attempt to retrieve successive pages until the number of records is
            less than `size`.  If the query is still incomplete after this many pages, just return
            what we've got.

        """
        params = {}
        if q is not None:
            params["q"] = q
        if sort is not None and sort in ["bestmatch", "mostrecent", "-bestmatch", "-mostrecent"]:
            params["sort"] = sort
        params["page"] = page
        params["size"] = size
        if allversions:
            params["allversions"] = ""

        url = "{0}api/records".format(self.base_url)
        r = self.session.get(url, params=params)
        if r.status_code != 200:
            print("An unknown error occurred when trying to access {0}.".format(url))
            print('The search parameters were "{0}"'.format(params))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

        json = r.json()
        if len(json) == size:
            page += 1
            if page > max_pages:
                print("Search is not yet complete after {0} pages; returning with what we have.".format(max_pages))
                return json  # Note: This will percolate back up the recursion to include other results
            return json + self.search(
                q=q, sort=sort, page=page, size=size, allversions=allversions, max_pages=max_pages
            )

        return json

    def delete_untitled_empty_deposits(self):
        deleted_deposits = 0
        deposits = self.search()
        for d in deposits:
            try:
                if d["title"] == "":
                    d = self.deposit(d["id"], ignore_deletion=True)
                    if not d.files:
                        d.delete_deposit(confirmed=True)
                        deleted_deposits += 1
            except:
                pass
        print("Deleted {0} deposits".format(deleted_deposits))

    def discard_all_drafts(self):
        discarded_drafts = 0
        deposits = self.search()
        for d in deposits:
            try:
                if d["state"] == "inprogress":
                    d = self.deposit(d["id"], ignore_deletion=True)
                    d.discard()
                    discarded_drafts += 1
            except:
                pass
        print("Discarded {0} drafts".format(discarded_drafts))

    def awaiting_approval(self, community_id):
        """List all records awaiting approval for the given community"""
        url = "{0}/api/records/?q=provisional_communities:{1}".format(self.base_url, community_id)
        r = self.session.get(url)
        if r.status_code != 200:
            print("Unable to find any records for community {0}.".format(community_id))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r.json()

    def community_curate_accept(self, community_id, record_id):
        """Accept a record into the community"""
        url = "{0}/communities/{1}/curaterecord/".format(self.base_url, community_id)
        data = {"recid": int(record_id), "action": "accept"}
        r = self.session.post(url, json=data)
        if r.status_code != 200:
            print(
                "Unable to accept record id {0} into community {1}; status code={2}.".format(
                    record_id, community_id, r.status_code
                )
            )
            try:
                r_json = r.json()
                print("Response JSON:")
                print(r_json)
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r.json()

    def total_deposit_size(self, deposition_id=None, human_readable=True):
        import math

        def convert_size(size_bytes):
            if size_bytes == 0:
                return "0B"
            size_name = ("B  ", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
            i = int(math.floor(math.log(size_bytes, 1024)))
            p = math.pow(1024, i)
            s = round(size_bytes / p, 3)
            return "{0:>8.3f} {1}".format(s, size_name[i])

        if deposition_id is None:
            depositions = self.search()
        else:
            depositions = [self.deposit(deposition_id, ignore_deletion=True).representation]
        total_size = 0
        for deposition in depositions:
            id = deposition["id"]
            d = self.deposit(id, ignore_deletion=True)
            d_total_size = sum([f["filesize"] for f in d.files])
            print('{1} in "{2}" (CaltechDATA ID {0})'.format(id, convert_size(d_total_size), d.title))
            total_size += d_total_size
        print("{0} in {1} deposits".format(convert_size(total_size), len(depositions)))
        if human_readable:
            return convert_size(total_size)  # Note: the return type will be str
        else:
            return total_size  # Note: the return type will be int



"""Search the INSPIRE database

Documentation can be found here: <https://github.com/inspirehep/rest-api-doc>

"""

import warnings

api_url = "https://inspirehep.net/api/{record_type}"


def inspire2doi(inspire_bibtex_key, raise_exceptions=False):
    try:
        result = query(inspire_bibtex_key, fields="dois.value,arxiv_eprints")
    except KeyboardInterrupt:
        raise
    except Exception as e:
        warnings.warn(f"Error querying for INSPIRE key {inspire_bibtex_key}")
        if raise_exceptions:
            raise e
    else:
        if not result:
            warnings.warn(f"No entry found for INSPIRE key {inspire_bibtex_key}")
            if raise_exceptions:
                raise ValueError(f"No entry found for INSPIRE key {inspire_bibtex_key}")
        else:
            try:
                return result[0]["metadata"]["dois"][0]["value"]
            except Exception as e:
                # DOI not present; INSPIRE only has an arXiv number, so we'll convert to an arXiv DOI
                try:
                    eprint_value = result[0]["metadata"]["arxiv_eprints"][0]["value"]
                    # https://info-arxiv-org.proxy.library.cornell.edu/help/arxiv_identifier.html
                    # says that identifiers starting with 9107 to 0703 need their category included
                    if eprint_value.startswith("9") or int(eprint_value[:4])<704:
                        category = result[0]["metadata"]["arxiv_eprints"][0]["categories"][0]
                        return f"10.48550/arXiv.{category}/{eprint_value}"
                    else:
                        return f"10.48550/arXiv.{eprint_value}"
                except Exception as e:
                    warnings.warn(f"Unexpected result format for INSPIRE key {inspire_bibtex_key}: \n{result}")
                    if raise_exceptions:
                        raise e


def query(
    query,
    sort="mostrecent",
    page=1,
    size=1000,
    fields=None,
    page_limit=10,
    record_type="literature"
):
    """Fetch records from the INSPIRE API.

    Parameters
    ----------
    query : str
        The search query.  For details, see
        https://github.com/inspirehep/rest-api-doc/tree/master#search-query
    fields : str, optional
        The fields to retrieve from INSPIRE.  This can be a
        comma-separated list of field names.  For example, to
        retrieve the list of DOIs: "dois.value".  Other options at
        https://inspire-schemas.readthedocs.io/en/latest/schemas/elements/reference.html
        Note that other fields will also be returned for each record,
        even when you just ask for one — including "id", "links",
        "metadata", "created", and "updated".  The default value of
        None will return all fields.
    sort : str, optional
        The sorting order. Default is "mostrecent".
    size : int, optional
        The number of records per page. Default is 1000.
    page : int, optional
        The starting page number. Default is 1.
    page_limit : int, optional
        Optional limit to the number of pages to collect.
    record_type : str, optional
        The type of record to fetch (e.g., "literature"). Default is
        "literature".  Other options can be found here:
        https://github.com/inspirehep/rest-api-doc/tree/master#obtaining-a-record
    
    Returns
    -------
    list
        The JSON response from the API.

    Raises
    ------
    requests.exceptions.HTTPError :
        If the HTTP request to INSPIRE failed

    """
    import sys
    import requests
    from requests.adapters import HTTPAdapter
    from urllib3.util.retry import Retry

    session = requests.Session()
    collected_results = []

    ## Retry automatically on certain types of errors
    retry = Retry(
        total=10,
        backoff_factor=0.1,
        status_forcelist=[
            413,  # Request Entity Too Large
            429,  # Too Many Requests
            500,  # Internal Server Error
            502,  # Bad Gateway
            503,  # Service Unavailable
            504,  # Gateway Timeout
        ],
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)

    url = api_url.format(record_type=record_type)
    params = {
        "q": query,
        "sort": sort,
        "size": size,
        "page": page,
    }
    if fields:
        params["fields"] = fields

    while params["page"] - page <= page_limit:
        response = session.get(url, params=params)
        if response.status_code != 200:
            print(f"An error occurred when trying to access <{api_url}>.", file=sys.stderr)
            try:
                print(response.json(), file=sys.stderr)
            except:
                pass
            response.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

        try:
            json_response = response.json()
        except ValueError:
            print(f"Response from {url} does not contain valid JSON; returning early.")
            return collected_results

        new_hits = json_response.get("hits", {}).get("hits", [])
        if not new_hits:
            break
        collected_results.extend(new_hits)

        params["page"] += 1

    return collected_results



import numpy as np
from . import jit


def transition_function(x, x0, x1, y0=0, y1=1, msquared=3):
    """Evaluate a C^∞ function transitioning from `y0` to `y1` in the range (`x0`, `x1`).

    This is a special type of [sigmoid
    function](https://en.wikipedia.org/wiki/Sigmoid_function) in which
    the transition region has finite size (between `x0` and `x1`), the
    function is constant outside of that region (equal to `y0` for
    values of `x <= x0` and `y1` for values of `x >= x1`), but the
    function is monotonic and infinitely differentiable everywhere —
    including *at* `x0` and `x1`.

    This particular function might be more descriptively called the
    "compactified tanh" function, because the nontrivial part is
    essentially the tanh function with its domain `(-∞,∞)` remapped —
    or compactified — to `(x0,x1)`.  This compactification could be
    applied to any sigmoid function; here we use the standard tanh
    function.

    The parameter `msquared` controls the abruptness of the
    transition.  The value `m` is the "fractional" slope of the
    function at the midpoint — that is, the slope is `m * (y1 - y0) /
    (x1 - x0)`.  The default value of `msquared = 3` corresponds to
    the most gradual transition with only one inflection point.
    Larger values will be more abrupt; smaller values will have
    multiple inflection points.  The least extreme second derivatives
    are obtained when `msquared` is approximately 4.0242806285.

    Parameters
    ----------
    x : array-like
        Input values.
    x0 : float
        Lower bound of the transition range.
    x1 : float
        Upper bound of the transition range.
    y0 : float, optional
        Function value at and below `x0`. Defaults to 0.
    y1 : float, optional
        Function value at and above `x1`. Defaults to 1.
    msquared : float, optional
        Controls the abruptness of the transition. Defaults to 3.

    Returns
    -------
    array-like
        The function values at `x`.
    """
    y = np.empty_like(x, dtype=np.promote_types(x.dtype, type(y1)))
    return transition_function_inplace(y, x, x0, x1, y0, y1, msquared)


@jit
def transition_function_inplace(y, x, x0, x1, y0=0, y1=1, msquared=3):
    """Evaluate `transition_function` in place on `y`
    
    See that function's docstring for details.
    """
    assert x0 <= x1
    m = np.sqrt(msquared)
    Δy = (y1 - y0) / 2
    Δx = (x1 - x0) / 2
    ȳ = (y0 + y1) / 2
    x̄ = (x0 + x1) / 2
    for i in range(x.size):
        x̂ = (x[i] - x̄) / Δx
        # Note that the conditions after the `or`s in the following
        # conditions will be *almost* always equivalent to the first
        # conditions, but they are necessary to protect against
        # roundoff errors.
        if x̂ <= -1 or x[i] <= x0:
            y[i] = y0
        elif x̂ >= 1 or x[i] >= x1:
            y[i] = y1
        else:
            y[i] = ȳ + Δy * np.tanh(m * x̂ / (1 - x̂**2))
    return y


def transition_function_derivative(x, x0, x1, y0=0, y1=1, msquared=3):
    """Evaluate the derivative of `transition_function`.

    See the `transition_function` docstring for details about the
    parameters.
    """
    y = np.empty_like(x, dtype=np.promote_types(x.dtype, type(y1)))
    return transition_function_derivative_inplace(y, x, x0, x1, y0, y1, msquared)


@jit
def transition_function_derivative_inplace(yprime, x, x0, x1, y0=0, y1=1, msquared=3):
    """Evaluate `transition_function_derivative` in place on `yprime`
    
    See that function's docstring for details.
    """
    assert x0 <= x1
    m = np.sqrt(msquared)
    Δy = (y1 - y0) / 2
    Δx = (x1 - x0) / 2
    x̄ = (x0 + x1) / 2
    for i in range(x.size):
        x̂ = (x[i] - x̄) / Δx
        if x̂ <= -1 or x[i] <= x0 or x̂ >= 1 or x[i] >= x1:
            yprime[i] = 0
        else:
            yprime[i] = Δy * m * (1 + x̂**2) / (Δx * (1 - x̂**2)**2 * np.cosh(m * x̂ / (1 - x̂**2))**2)
    return yprime


def transition_to_constant(f, t, t1, t2):
    """Smoothly transition from the function to a constant.

    This works (implicitly) by multiplying the derivative of `f` with
    the transition function going from 1 at `t<=t1` to 0 at `t>=t2`,
    and then integrating.  Using integration by parts, this simplifies
    to multiplying `f` itself by the transition function, and then
    subtracting the integral of `f` times the derivative of the
    transition function.  This integral is effectively restricted to
    the region (t1, t2).  Note that the final value (after t2) will
    depend on the precise values of `t1` and `t2`, and the behavior of
    `f` in between.

    Parameters
    ----------
    f : array_like
        Array corresponding to the following `t` parameter.  If this
        array is multi-dimensional, the first dimension must be the
        same size as `t`.
    t : array_like
        One-dimensional monotonic array of floats.
    t1 : float
        Value before which the output will equal `f`.
    t2 : float
        Value after which the output will be constant.

    """
    fprime = np.asanyarray(f).copy()
    transition_to_constant_inplace(fprime, t, t1, t2)
    return fprime


def transition_to_constant_inplace(f, t, t1, t2):
    """Evaluate `transition_to_constant` in place on `f`
    
    See that function's docstring for details.
    """
    from scipy.interpolate import CubicSpline
    assert f.shape[0] == t.size
    i1, i2 = np.searchsorted(t, [t1, t2])
    τ = transition_function(t, t1, t2, y0=1, y1=0)
    τ̇ = transition_function_derivative(t, t1, t2, y0=1, y1=0)
    # Note that numpy's broadcasting rules add extra dimensions on the
    # left, which means we need to move the time dimension to the
    # right to multiply `f` by `t`, then move it back.  This is the
    # reason for the double transposes.
    intfτ̇ = CubicSpline(
        t[i1:i2], (f[i1:i2].T * τ̇[i1:i2]).T
    ).antiderivative()(t[i1:i2])
    f[i1:i2] = (f[i1:i2].T * τ[i1:i2]).T - intfτ̇
    f[i2:] = f[i2 - 1]
    return f


def bump_function(x, x0, x1, x2, x3, y0=0, y12=1, y3=None, m01squared=3, m23squared=None):
    """Evaluate a C^∞ "bump" function.

    This is essentially the same as `transition_function`, but with
    *two* transitions.  See that function's documentation for details
    about the meanings of the parameters.

    In short, the function's value is `y0` for `x <= x0`, smoothly
    transitions from there to `y12` on `x0 < x < x1`, stays at `y12`
    for `x1 <= x <= x2`, smoothly transitions from there to `y3` for
    `x2 < x < x3`, and stays at `y3` for `x >= x3`.

    Parameters
    ----------
    x : array-like
        Input values.
    x0 : float
        Lower bound of the first transition range.
    x1 : float
        Upper bound of the first transition range.
    x2 : float
        Lower bound of the second transition range.
    x3 : float
        Upper bound of the second transition range.
    y0 : float, optional
        Function value at and below `x0`. Defaults to 0.
    y12 : float, optional
        Function value in the first transition range. Defaults to 1.
    y3 : float, optional
        Function value at and above `x3`. Defaults to 0.
    m01squared : float, optional
        Controls the abruptness of the first transition. Defaults to 3.
    m23squared : float, optional
        Controls the abruptness of the second transition. Defaults to `m01squared`.

    Returns
    -------
    array-like
        The function values at `x`.
    """
    y3 = y3 if y3 is not None else y0
    y = np.empty_like(x, dtype=np.promote_types(x.dtype, type(y3)))
    return bump_function_inplace(y, x, x0, x1, x2, x3, y0, y12, y3, m01squared, m23squared)


@jit
def bump_function_inplace(y, x, x0, x1, x2, x3, y0=0, y12=1, y3=0, m01squared=3, m23squared=None):
    """Evaluate `bump_function` in place on `y`
    
    See that function's docstring for details.
    """
    assert x0 <= x1 <= x2 <= x3
    m01 = np.sqrt(m01squared)
    m23 = np.sqrt(m23squared if m23squared is not None else m01squared)
    Δy01 = (y12 - y0) / 2
    Δy23 = (y3 - y12) / 2
    Δx01 = (x1 - x0) / 2
    Δx23 = (x3 - x2) / 2
    ȳ01 = (y0 + y12) / 2
    ȳ23 = (y12 + y3) / 2
    x̄01 = (x0 + x1) / 2
    x̄23 = (x2 + x3) / 2
    for i in range(x.size):
        x̂01 = (x[i] - x̄01) / Δx01
        x̂23 = (x[i] - x̄23) / Δx23
        # Note that the conditions after the `or`s in the following
        # conditions will be *almost* always equivalent to the first
        # conditions, but they are necessary to protect against
        # roundoff errors.
        if x̂01 <= -1 or x[i] <= x0:
            y[i] = y0
        elif (x̂01 >= 1 or x[i] >= x1) and (x̂23 <= -1 or x[i] <= x2):
            y[i] = y12
        elif x̂23 >= 1 or x[i] >= x3:
            y[i] = y3
        elif -1 < x̂01 < 1:
            y[i] = ȳ01 + Δy01 * np.tanh(m01 * x̂01 / (1 - x̂01**2))
        else:  # -1 < x̂23 < 1
            y[i] = ȳ23 + Δy23 * np.tanh(m23 * x̂23 / (1 - x̂23**2))
    return y



import numpy as np
import quaternionic
from .. import WaveformModes
from . import PostNewtonian

def PNWaveform(
    Approximant, delta, chi1_i, chi2_i, Omega_orb_i, *,
    Omega_orb_0=None, R_frame_i=None, MinStepsPerOrbit=32,
    PNWaveformModeOrder=4.0, PNOrbitalEvolutionOrder=4.0,
    inertial=False, dt=0.0, quiet=True,
    ell_min=2, ell_max=8, Lambda1=0, Lambda2=0,
    **kwargs
):
    """Generate a PN waveform.

    The return value is an `sxs.WaveformModes` object with the
    following additional fields:

        - `M1` (array): The primary mass as a function of time.
        - `M2` (array): The secondary mass as a function of time.
        - `chi1` (array): The primary spin as a function of time.
        - `chi2` (array): The secondary spin as a function of time.
        - `frame` (array): The quaternionic frame as a function of
          time.
        - `v` (array): The orbital velocity as a function of time.
        - `orbital_phase` (array): The orbital phase as a function of
          time.
    
    This is a wrapper around the Julia function
    `PostNewtonian.GWFrames.PNWaveform`.  See [the Julia
    documentation](https://moble.github.io/PostNewtonian.jl/dev/interface/gwframes/#PostNewtonian.GWFrames.PNWaveform)
    for details on the optional keyword arguments.
    
    """
    Omega_orb_0 = Omega_orb_0 or Omega_orb_i
    R_frame_i = R_frame_i or [1.0]

    # Integrate the orbital dynamics
    w1 = PostNewtonian.GWFrames.PNWaveform(
        Approximant, delta, chi1_i, chi2_i, Omega_orb_i,
        Omega_orb_0=Omega_orb_0, R_frame_i=R_frame_i,
        MinStepsPerOrbit=MinStepsPerOrbit,
        PNWaveformModeOrder=PNWaveformModeOrder,
        PNOrbitalEvolutionOrder=PNOrbitalEvolutionOrder,
        inertial=inertial, dt=dt, quiet=quiet,
        ell_min=ell_min, ell_max=ell_max, Lambda1=Lambda1, Lambda2=Lambda2,
        **kwargs
    )

    coorbital_frame = quaternionic.array(w1.frame.to_numpy())
    frame = np.array([quaternionic.one]) if inertial else coorbital_frame
    frame_type = "inertial" if inertial else "coorbital"

    w = WaveformModes(
        w1.data.to_numpy(),
        time=w1.t.to_numpy(),
        modes_axis=1,
        ell_min=ell_min,
        ell_max=ell_max,
        spin_weight=-2,
        frame=frame,
        frame_type=frame_type,
        data_type="h",
        M1=w1.M1.to_numpy(),
        M2=w1.M2.to_numpy(),
        chi1=w1.chi1.to_numpy(),
        chi2=w1.chi2.to_numpy(),
        coorbital_frame=coorbital_frame,
        v=w1.v.to_numpy(),
        orbital_phase=w1.Phi.to_numpy(),
    )

    # Allow the extra fields to be accessed naturally
    w.M1 = w._metadata["M1"]
    w.M2 = w._metadata["M2"]
    w.chi1 = w._metadata["chi1"]
    w.chi2 = w._metadata["chi2"]
    w.coorbital_frame = w._metadata["coorbital_frame"]
    w.v = w._metadata["v"]
    w.orbital_phase = w._metadata["orbital_phase"]

    return w



format_description = """New catalog format

{
    "sxs_format": "catalog_v2"
    "modified": "<YYYY-MM-DDThh:mm:ss.ssssss>",  # UTC time of last-modified record in this file

    "records": {  # Includes *all* versions
        "<sxs_id_versioned>": {  # SXS:(BBH|BHNS|NSNS):[0-9]{4,}v[0-9]{2,}
            "title": "<title>"  # Same as ["metadata"]["title"],
            "id": <id>,  # ~7-digit integer uniquely identifying this record
            "conceptrecid": "<conceptrecid>",  # ~7-digit integer (as string) collectively identifying all versions of this record
            "created": "<YYYY-MM-DDThh:mm:ss.ssssss>",  # UTC time of creation of this record on Zenodo
            "modified": "<YYYY-MM-DDThh:mm:ss.ssssss>",  # (UTC) Last modification of this record (possibly just Zenodo metadata modified)
            "bucket": "https://zenodo.org/api/files/<uuid>",  # Base URL for file uploads and downloads,
            "files": [
                {
                    "checksum": "<checksum>",  # MD5 checksum of file on Zenodo
                    "filename": "<filename>",  # Name of file; may contain slashes denoting directories
                    "filesize": <filesize>,  # Number of bytes in the file
                    "id": "<fileid>",  # A standard UUID (hexadecimal with characters in the pattern 8-4-4-4-12)
                },
                ...  # Other file descriptions in the order in which they were uploaded (not necessarily a meaningful order)
            ]
        },
        ...
    },

    "simulations": {  # Physical data (masses, spins, etc.) for all available SXS simulations, in the most recent version
        "<sxs_id>": {  # The SXS ID is a string like SXS:BHNS:0001 or SXS:BBH:1234
            "latest_version": "<sxs_id_versioned>",  # Entry in "records" containing the most recent version of this simulation's data
            "url": "<URL>",  # The URL of the Zenodo "conceptdoi" link, which *resolves to* the most-recent version
            #
            # NOTE: All of the following may be absent if this simulation is closed-access, or simply does not have metadata.
            #
            # Variable content describing (mostly) physical parameters of the system.  It's basically a
            # python-compatible version of the information contained in "metadata.txt" from the
            # highest-resolution run in the most-recent version of this simulation.  That file is meant to
            # be more-or-less as suggested in <https://arxiv.org/abs/0709.0093>.  The conversion to a
            # python-compatible format means that keys like "simulation-name" have had hyphens replaced by
            # underscores so that they can be used as variable names in python and any other sane language
            # (with apologies to Lisp).  As far as possible, values that are just strings in that file
            # have been converted into the relevant types -- like numbers, integers, and arrays.  Note
            # that some keys like eccentricity are sometimes numbers and sometimes the string "<number"
            # (meaning that the eccentricity is less than the number), which is necessarily a string.
            #
            # Below are just the first few keys that *may* be present.  Note that closed-access
            # simulations will have empty dictionaries here.
            #
            "simulation_name": "<directory_name>",  # This may be distinctly uninformative
            "alternative_names": "<some ugly thing>, ..., <sxs_id>",  # This may be a list of strings
            "initial_data_type": "<type>",  # Something like "BBH_CFMS"
            "object_types": "<type>",  # Currently "BHBH", "BHNS", or "NSNS"
            "number_of_orbits": <number>,  # This is a float, rather than an integer
            "reference_mass_ratio": <q>,  # Usually greater than 1 (exceptions are due to junk radiation)
            "reference_chi_eff": <chi_eff>,  # Dimensionless effective spin quantity
            "reference_chi1_perp": <chi1_perp>,  # Magnitude of component of chi1 orthogonal to "reference_orbital_frequency"
            "reference_chi2_perp": <chi2_perp>,  # Magnitude of component of chi2 orthogonal to "reference_orbital_frequency"
            "reference_mass1": <m2>,
            "reference_mass2": <m1>,
            "reference_dimensionless_spin1": [
                <chi1_x>,
                <chi1_y>,
                <chi1_z>
            ],
            "reference_dimensionless_spin2": [
                <chi2_x>,
                <chi2_y>,
                <chi2_z>
            ],
            "reference_eccentricity": <eccentricity>,  # A float or possibly a string containing "<" and a float
            "reference_orbital_frequency": [
                <omega_x>,
                <omega_y>,
                <omega_z>
            ],
            "reference_time": <time>,
            ...
        },
        ...
    }

}


"""



"""Container interface to the catalog of SXS simulations"""

import functools
import collections


class Simulations(collections.OrderedDict):
    """Interface to the catalog of SXS simulations
    
    Creation
    --------
    You probably don't need to create this object yourself.  The
    easiest way to create this object is just to use the `sxs.load`
    function:

    ```python
    import sxs

    simulations = sxs.load("simulations")
    ```
    """
    last_modified_url = "https://api.github.com/repos/sxs-collaboration/sxs/contents/simulations.json?ref=simulations"
    url = "https://github.com/sxs-collaboration/sxs/raw/simulations/simulations.json"

    def __init__(self, sims):
        """Initialize the Simulations dictionary

        Note that the constructor is not generally useful from outside
        this class.  See `Simulations.load` for a more useful
        initialization function, or simply call
        `sxs.load("simulations")`.

        """
        from .. import Metadata
        super(Simulations, self).__init__(
            (k, Metadata(sims[k])) for k in sorted(sims)
        )

    @classmethod
    def remote_timestamp(cls, download):
        import requests
        from datetime import datetime, timezone
        if not download:
            return datetime.min.replace(tzinfo=timezone.utc)
        failed = False
        try:
            response = requests.head(
                Simulations.last_modified_url,
                headers={"X-GitHub-Api-Version": "2022-11-28"},
            )
            if response.status_code != 200 or "Last-Modified" not in response.headers:
                failed = True
            else:
                remote_timestamp = datetime.strptime(
                    response.headers["Last-Modified"], "%a, %d %b %Y %H:%M:%S GMT"
                ).replace(tzinfo=timezone.utc)
        except:
            failed = True
        if failed:
            print(
                f"Failed to get the remote timestamp from <{Simulations.last_modified_url}>.\n"
                + "Assuming it is old."
            )
            return datetime.min.replace(tzinfo=timezone.utc)
        return remote_timestamp

    @classmethod
    @functools.lru_cache()
    def load(cls, download=None):
        """Load the catalog of SXS simulations

        Note that — unlike most SXS data files — the simulations file is updated
        frequently.  As a result, this function — unlike the loading functions for most
        SXS data files — will download the simulations by default each time it is called.
        However, also note that this function is itself cached, meaning that the same
        dict will be returned on each call in a given python session.  If you want to
        avoid that behavior, use `Simulations.reload`.

        Parameters
        ----------
        download : {None, bool}, optional
            If False, this function will look for the simulations in the sxs cache and
            raise an error if it is not found.  If True, this function will download
            the simulations and raise an error if the download fails.  If None (the
            default), it will try to download the file, warn but fall back to the cache
            if that fails, and only raise an error if the simulations is not found in the
            cache.  Note that this ignores the sxs configuration file entirely.

        See Also
        --------
        sxs.sxs_directory : Locate cache directory
        Simulations.reload : Avoid caching the result of this function

        """
        from datetime import datetime, timezone
        import json
        import zipfile
        from .. import sxs_directory, read_config
        from ..utilities import download_file

        progress = read_config("download_progress", True)

        remote_timestamp = cls.remote_timestamp(download is not False)  # Test for literal `False`

        cache_path = sxs_directory("cache") / "simulations.zip"

        if cache_path.exists():
            local_timestamp = datetime.fromtimestamp(cache_path.stat().st_mtime, timezone.utc)
        elif download is False:
            raise ValueError(f"Simulations not found in '{cache_path}' and downloading was turned off")
        else:
            local_timestamp = datetime.min.replace(tzinfo=timezone.utc)

        download_failed = False
        if (download or download is None) and remote_timestamp > local_timestamp:
            # 1. Download the full json file (zipped in flight, but auto-decompressed on arrival)
            # 2. Zip to a temporary file (using bzip2, which is better than the in-flight compression)
            # 3. Replace the original simulations.zip with the temporary zip file
            # 4. Remove the full json file
            # 5. Make sure the temporary zip file is gone too
            temp_json = cache_path.with_suffix(".temp.json")
            temp_zip = cache_path.with_suffix(".temp.zip")
            try:
                try:
                    download_file(cls.url, temp_json, progress=progress, if_newer=False)
                except Exception as e:
                    if download:
                        raise RuntimeError(f"Failed to download '{cls.url}'; try setting `download=False`") from e
                    download_failed = e  # We'll try the cache
                else:
                    if temp_json.exists():
                        with zipfile.ZipFile(temp_zip, "w", compression=zipfile.ZIP_BZIP2) as simulations_zip:
                            simulations_zip.write(temp_json, arcname="simulations.json")
                        temp_zip.replace(cache_path)
            finally:
                temp_json.unlink(missing_ok=True)
                temp_zip.unlink(missing_ok=True)

        if not cache_path.exists():
            if download is False:  # Test if it literally *is* False, rather than just casts to False
                raise ValueError(f"The simulations file was not found in '{cache_path}', and downloading was turned off")
            elif download_failed:
                raise ValueError(f"Simulations not found in '{cache_path}' and download failed") from download_failed
            else:
                raise ValueError(f"Simulations not found in '{cache_path}' for unknown reasons")

        try:
            with zipfile.ZipFile(cache_path, "r") as simulations_zip:
                try:
                    with simulations_zip.open("simulations.json") as simulations_json:
                        try:
                            simulations = json.load(simulations_json)
                        except Exception as e:
                            raise ValueError(f"Failed to parse 'simulations.json' in '{cache_path}'") from e
                except Exception as e:
                    raise ValueError(f"Failed to open 'simulations.json' in '{cache_path}'") from e
        except Exception as e:
            raise ValueError(f"Failed to open '{cache_path}' as a ZIP file") from e

        return cls(simulations)

    @classmethod
    def reload(cls, download=True):
        """Reload the catalog of SXS simulations, without caching

        Clears the cache of `Simulations.load` and returns the result of calling it again.
        Note that in this function, the default value of `download` is `True`, rather
        than `None` as in `Simulations.load` — though both behaviors are available.

        Parameters
        ----------
        download : {None, bool}, optional
            If False, this function will look for the simulations in the sxs cache and
            raise an error if it is not found.  If True (the default), this function
            will download the simulations and raise an error if the download fails.  If
            None (the default), it will try to download the file, warn but fall back to
            the cache if that fails, and only raise an error if the simulations is not
            found in the cache.  Note that this ignores the sxs configuration file
            entirely.

        See Also
        --------
        sxs.sxs_directory : Locate cache directory
        Simulations.load : Caching version of this function

        """
        cls.load.cache_clear()
        return cls.load(download=download)

    @property
    def dataframe(self):
        """Create pandas.DataFrame containing metadata for all
        simulations

        Note that `pandas` is the standard Python interface for
        heterogeneous data tables, like the one we have here.  This
        interface allows for more convenient slicing and querying of
        data than the list of `dict`s provided by the `Simulations`
        object.
        
        This can also be a more convenient way to access the metadata
        because the raw metadata has missing keys and mixed formats.
        Iif a key is missing from the metadata for a particular key,
        the dataframe will just have a `NaN` in that entry, rather
        than raising an exception.  Other keys may have unexpected
        entries — such as the `"reference_eccentricity"` field, which
        is *usually* a float but may be a string like "<0.0001" if the
        eccentricity is not known precisely, but is only bounded.  The
        dataframe introduces a new column called
        `"reference_eccentricity_bound"` that is always a float giving
        an upper bound on the eccentricity.

        See the `pandas` documentation for more information on how to
        use the resulting dataframe, or the `Simulations` tutorial for
        examples.
        
        """
        import numpy as np
        import pandas as pd

        if hasattr(self, "_dataframe"):
            return self._dataframe

        simulations = pd.DataFrame.from_dict(self, orient="index")

        def floater(x):
            try:
                f = float(x)
            except:
                f = np.nan
            return f

        def floaterbound(x):
            try:
                f = float(x)
            except:
                try:
                    f = float(x.replace("<", ""))
                except:
                    f = np.nan
            return f

        def norm(x):
            try:
                n = np.linalg.norm(x)
            except:
                n = np.nan
            return n

        def three_vec(x):
            try:
                a = np.array(x, dtype=float)
                if a.shape != (3,):
                    raise ValueError("Don't understand input as a three-vector")
            except:
                a = np.array([np.nan, np.nan, np.nan])
            return a

        def datetime_from_string(x):
            try:
                dt = pd.to_datetime(x).tz_convert("UTC")
            except:
                dt = pd.to_datetime("1970-1-1").tz_localize("UTC")
            return dt

        sims_df = pd.concat((
            simulations["reference_time"].map(floater),
            simulations["reference_mass_ratio"].map(floater),
            simulations["reference_dimensionless_spin1"].map(three_vec),
            simulations["reference_dimensionless_spin1"].map(norm).rename("reference_dimensionless_spin1_mag"),
            simulations["reference_dimensionless_spin2"].map(three_vec),
            simulations["reference_dimensionless_spin2"].map(norm).rename("reference_dimensionless_spin2_mag"),
            simulations["reference_chi_eff"].map(floater),
            simulations["reference_chi1_perp"].map(floater),
            simulations["reference_chi2_perp"].map(floater),
            simulations["reference_eccentricity"].map(floater),
            simulations["reference_eccentricity"].map(floaterbound).rename("reference_eccentricity_bound"),
            simulations["reference_mean_anomaly"].map(floater),
            simulations["reference_orbital_frequency"].map(three_vec),
            simulations["reference_orbital_frequency"].map(norm).rename("reference_orbital_frequency_mag"),
            (
                simulations["reference_position1"].map(three_vec)
                -simulations["reference_position2"].map(three_vec)
            ).map(norm).rename("reference_separation"),
            simulations["reference_position1"].map(three_vec),
            simulations["reference_position2"].map(three_vec),
            simulations["reference_mass1"].map(floater),
            simulations["reference_mass2"].map(floater),
            simulations["reference_dimensionless_spin1"].map(norm).rename("reference_chi1_mag"),
            simulations["reference_dimensionless_spin2"].map(norm).rename("reference_chi2_mag"),
            simulations["relaxation_time"].map(floater),
            #simulations["merger_time"].map(floater),
            simulations["common_horizon_time"].map(floater),
            simulations["remnant_mass"].map(floater),
            simulations["remnant_dimensionless_spin"].map(three_vec),
            simulations["remnant_dimensionless_spin"].map(norm).rename("remnant_dimensionless_spin_mag"),
            simulations["remnant_velocity"].map(three_vec),
            simulations["remnant_velocity"].map(norm).rename("remnant_velocity_mag"),
            #simulations["final_time"].map(floater),
            simulations["EOS"].fillna(simulations["eos"]),
            simulations["disk_mass"].map(floater),
            simulations["ejecta_mass"].map(floater),
            simulations["object_types"].astype("category"),
            simulations["initial_data_type"].astype("category"),
            simulations["initial_separation"].map(floater),
            simulations["initial_orbital_frequency"].map(floater),
            simulations["initial_adot"].map(floater),
            simulations["initial_ADM_energy"].map(floater),
            simulations["initial_ADM_linear_momentum"].map(three_vec),
            simulations["initial_ADM_linear_momentum"].map(norm).rename("initial_ADM_linear_momentum_mag"),
            simulations["initial_ADM_angular_momentum"].map(three_vec),
            simulations["initial_ADM_angular_momentum"].map(norm).rename("initial_ADM_angular_momentum_mag"),
            simulations["initial_mass1"].map(floater),
            simulations["initial_mass2"].map(floater),
            simulations["initial_mass_ratio"].map(floater),
            simulations["initial_dimensionless_spin1"].map(three_vec),
            simulations["initial_dimensionless_spin1"].map(norm).rename("initial_dimensionless_spin1_mag"),
            simulations["initial_dimensionless_spin2"].map(three_vec),
            simulations["initial_dimensionless_spin2"].map(norm).rename("initial_dimensionless_spin2_mag"),
            simulations["initial_position1"].map(three_vec),
            simulations["initial_position2"].map(three_vec),
            #simulations["object1"].astype("category"),
            #simulations["object2"].astype("category"),
            # simulations["url"],
            #simulations["simulation_name"],
            #simulations["alternative_names"],
            # simulations["metadata_path"],
            # simulations["end_of_trajectory_time"].map(floater),
            # simulations["merger_time"].map(floater),
            simulations["number_of_orbits"].map(floater),
            simulations["superseded_by"],
            simulations["DOI_versions"],
            simulations["keywords"],
            simulations["date_link_earliest"].map(datetime_from_string),
            simulations["date_run_earliest"].map(datetime_from_string),
            simulations["date_run_latest"].map(datetime_from_string),
            simulations["date_postprocessing"].map(datetime_from_string),
        ), axis=1)

        sims_df.insert(0, "deprecated", (
            ~sims_df.superseded_by.isna()
            | sims_df["keywords"].map(lambda ks: "deprecated" in ks)
        ))

        # We have ignored the following fields present in the
        # simulations.json file (as of 2024-08-04), listed here with
        # the number of non-null entries:
        #
        # alternative_names                2778
        # point_of_contact_email           2778
        # authors_emails                   2776
        # simulation_bibtex_keys           2778
        # code_bibtex_keys                 2778
        # initial_data_bibtex_keys         2778
        # quasicircular_bibtex_keys        2778
        # metadata_version                 2778
        # spec_revisions                   2778
        # spells_revision                  2778
        # merger_time                         9
        # final_time                         12
        # reference_spin1                     2
        # reference_spin2                     1
        # nitial_spin1                        2
        # initial_spin2                       2
        # remnant_spin                        2
        # initial_mass_withspin2              2
        # end_of_trajectory_time              3

        self._dataframe = sims_df
        return sims_df

    table = dataframe



"""Various utilities used the sxs package"""

import functools
import numba

jit = functools.partial(numba.njit, cache=True)
vectorize = functools.partial(numba.vectorize, nopython=True, cache=True)
guvectorize = functools.partial(numba.guvectorize, nopython=True, cache=True)

# This is a generically reliable set of widths to feed to multishuffle when using 64-bit floats
default_shuffle_widths = (16, 4, 2, 2, 2) + (1,) * 22 + (4,) * 4
default_shuffle_widths_old = (8, 8, 4, 4, 4, 2,) + (1,) * 34

from . import url, inspire, monotonicity, decimation, lvcnr, references
from .downloads import download_file
from .bitwise import diff, xor, multishuffle
from .sxs_identifiers import (
    sxs_identifier_regex, sxs_identifier_re,
    lev_regex, lev_re,
    sxs_id_version_lev_regex, sxs_id_version_lev_re,
    sxs_id_version_lev_exact_regex, sxs_id_version_lev_exact_re,
    sxs_id, sxs_id_and_version,
    lev_number, simulation_title, sxs_id_to_url,
)
from .sxs_directories import (
    sxs_directory, read_config, write_config, sxs_path_to_system_path, cached_path
)
from .select import select_by_path_component
from .formats import file_format
from .pretty_print import fit_to_console
from .files import (
    md5checksum, lock_file_manager, find_simulation_directories, find_files
)
from .dicts import KeyPassingDict
from .smooth_functions import (
    transition_function, transition_function_inplace,
    transition_function_derivative, transition_function_derivative_inplace,
    transition_to_constant, transition_to_constant_inplace,
    bump_function, bump_function_inplace
)
from .inspire import inspire2doi


def version_info():
    """Find all relevant package versions

    This function attempts to import each of the various packages relevant to this
    package — such as numpy, quaternionic, etc. — and returns a dictionary mapping
    the package names to their version strings.  It attempts to import some
    packages — like spinsfast and scri — which are not required; if they are not
    found, they simply are not included in the result.

    """
    import importlib
    import sys
    info = {
        "python": sys.version,
    }
    modules = [
        "numpy", "scipy", "h5py", "numba",
        "quaternion", "spherical_functions", "spinsfast", "scri",
        "quaternionic", "spherical", "sxs"
    ]
    for module in modules:
        try:
            m = importlib.import_module(module)
            info[module] = m.__version__
        except Exception:
            pass
    return info


class SimpleVersion:
    """Basic versioning class

    This class uses a simple two-element version string, enabling printing,
    incrementing, and comparison.

    Note that this is just intended for use SXS data versions; for more general
    use, you probably want something like
    [`packaging.version`](https://packaging.pypa.io/en/latest/version.html).

    """
    def __init__(self, v):
        if "." not in v:
            v = f"1.{v}"
        self.major, self.minor = map(int, v.split("."))
    def __str__(self):
        return f"{self.major}.{self.minor}"
    def __repr__(self):
        return f"{self.major}.{self.minor}"
    def increment(self, level="minor"):
        if level=="major":
            self.major += 1
        elif level=="minor":
            self.minor += 1
        else:
            raise ValueError("""Can only increment major or minor levels, not "{level}".""")
        return self

    def __hash__(self):
        return hash(str(self))

    def __lt__(self, other):
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major < other.major or (self.major == other.major and self.minor < other.minor)

    def __le__(self, other):
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major <= other.major or (self.major == other.major and self.minor <= other.minor)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major == other.major and self.minor == other.minor

    def __ge__(self, other):
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major >= other.major or (self.major == other.major and self.minor >= other.minor)

    def __gt__(self, other):
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major > other.major or (self.major == other.major and self.minor > other.minor)

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, SimpleVersion):
            return NotImplemented
        return self.major != other.major or self.minor != other.minor



from pathlib import Path
from warnings import warn
from .. import doi_url, Metadata
from ..utilities import (
    sxs_id_and_version, lev_number, sxs_path_to_system_path,
    download_file, sxs_directory, read_config
)


def Simulation(location, *args, **kwargs):
    """Construct a Simulation object from a location string

    The location string should be an SXS ID with an optional version
    number, and possibly a Lev specification, as in
    "SXS:BBH:1234v2.0/Lev5".  The default version number is "v2.0",
    and the default Lev number is the highest available.  Note that
    the version number must be in the form "v2.0", not "2.0", and it
    must exist, or no files will be found to load the data from.

    The returned object will be either a `Simulation_v1` or
    `Simulation_v2` object, depending on the version number.
    Hopefully, most details of the two versions will be hidden from
    the user, so that the interface is identical.

    Note that some simulations are deprecated and/or superseded by
    other simulations.  By default, this function will raise an error
    if you try to load a deprecated or superseded simulation.  There
    are several ways to override this behavior:

    1. Pass `ignore_deprecation=True` to completely bypass even
        checking for deprecation or supersession.  No warnings or
        errors will be issued.
    2. Include an explicit version number in the `location`
        string, as in "SXS:BBH:0001v2.0".  A warning will be issued
        that the simulation is deprecated, but it will be loaded
        anyway.
    3. Pass `auto_supersede=True` to automatically load the
        superseding simulation, if there is only one.  Because no
        superseding simulation can be *precisely* the same as the
        deprecated one, there may be multiple superseding simulations
        that have very similar parameters, in which case an error will
        be raised and you must explicitly choose one.  If there is
        only one, a warning will be issued, but the superseding
        simulation will be loaded.

    Otherwise, a `ValueError` will be raised, with an explanation and
    suggestions on what you might want to do.

    Parameters
    ----------
    location : str
        The location string for the simulation.  This must include the
        SXS ID, but can also include the version and Lev number, as
        described above.

    Other parameters
    ----------------
    ignore_deprecation : bool
        If `True`, completely bypass checking for deprecation or
        supersession.  No warnings or errors will be issued.
    auto_supersede : bool
        If `True`, automatically load the superseding simulation, if
        there is only one.  If there are multiple superseding
        simulations, an error will be raised, and you must explicitly
        choose one.  If there is only one, a warning will still be
        issued, but the superseding simulation will be loaded.  Note
        that this can also be set in the configuration file with
        `sxs.write_config(auto_supersede=True)`.
    extrapolation : str
        The extrapolation order to use for the strain and Psi4 data.
        This is only relevant for versions 1 and 2 of the data format,
        both of which default to "N2".  Other options include "N3",
        "N4", and "Outer".  "Nx" refers to extrapolation by
        polynomials in 1/r with degree `x`, while "Outer" refers to
        data extracted at the outermost extraction radius but
        corrected for time-dilation and areal-radius effects.
    download : bool
        If `True`, download the information about the files from the
        Zenodo/CaltechDATA record.  If `False`, only use the file
        information that is already available (which will raise an
        error if the file information has not previously been
        downloaded).  If not present, use the value from the
        configuration file, defaulting to `True` if it is not
        configured.

    Returns
    -------
    simulation : SimulationBase
        A `Simulation_v1` or `Simulation_v2` object, depending on the
        version of the simulation data.

    Note that all remaining arguments (including keyword arguments)
    are passed on to the `SimulationBase`, `Simulation_v1`, and/or
    `Simulation_v2` constructors, though none currently recognize any
    arguments other than those listed above.
    
    """
    from .. import load

    # Extract the simulation ID, version, and Lev from the location string
    simulation_id, input_version = sxs_id_and_version(location)
    if not simulation_id:
        raise ValueError(f"Invalid SXS ID in '{simulation_id}'")
    input_lev_number = lev_number(location)  # Will be `None` if not present

    # Load the simulation catalog and check if simulation ID exists in the catalog
    simulations = load("simulations")
    if simulation_id not in simulations:
        raise ValueError(f"Simulation '{simulation_id}' not found in simulation catalog")

    # Attach metadata to this object
    metadata = Metadata(simulations[simulation_id])
    series = simulations.dataframe.loc[simulation_id]

    # Check if the specified version exists in the simulation catalog
    if input_version not in metadata.DOI_versions:
        raise ValueError(f"Version '{input_version}' not found in simulation catalog for '{simulation_id}'")

    # Set various pieces of information about the simulation
    version = input_version or max(metadata.DOI_versions)
    if not version.startswith("v"):
        raise ValueError(f"Invalid version string '{version}'")
    sxs_id_stem = simulation_id
    sxs_id = f"{sxs_id_stem}{version}"
    url = f"{doi_url}{sxs_id}"

    # Deal with "superseded_by" field, or "deprecated" keyword in the metadata
    deprecated = ("deprecated" in metadata.get("keywords", []) or metadata.get("superseded_by", False))
    if not kwargs.get("ignore_deprecation", False):
        auto_supersede = kwargs.get("auto_supersede", read_config("auto_supersede", False))
        if (
            input_version
            and not auto_supersede
            and deprecated
        ):
            message = ("\n"
                + f"Simulation '{sxs_id_stem}' is deprecated and/or superseded.\n"
                + "Normally, this simulation should no longer be used, but you\n"
                + f"explicitly requested version '{input_version}', so it is being used.\n"
            )
            warn(message)
        else:
            if "superseded_by" in metadata:
                superseded_by = metadata["superseded_by"]
                if auto_supersede and isinstance(superseded_by, list):
                    raise ValueError(
                        f"`auto_supersede` is enabled, but simulation '{sxs_id}' is\n"
                        + "superseded by multiple simulations.  You must choose one\n"
                        + "explicitly from the list:\n"
                        + "\n".join(f"  {s}" for s in superseded_by)
                        + "\nAlternatively, you could pass `ignore_deprecation=True` or\n"
                        + "specify a version to load this waveform anyway."
                    )
                elif auto_supersede and isinstance(superseded_by, str):
                    message = f"\nSimulation '{sxs_id}' is being automatically superseded by '{superseded_by}'."
                    warn(message)
                    new_location = f"{superseded_by}{input_version}"
                    if input_lev_number:
                        new_location += f"/Lev{input_lev_number}"
                    return Simulation(new_location, *args, **kwargs)
                elif isinstance(superseded_by, list):
                    raise ValueError(
                        f"Simulation '{sxs_id}' is superseded by multiple simulations.\n"
                        + "Even if you enable `auto_supersede`, with multiple options, you\n"
                        + "must choose one explicitly from the list:\n"
                        + "\n".join(f"  {s}" for s in superseded_by)
                        + "\nAlternatively, you could pass `ignore_deprecation=True` or\n"
                        + "specify a version to load this waveform anyway."
                    )
                elif isinstance(superseded_by, str):
                    raise ValueError(
                        f"Simulation '{sxs_id}' is superseded by '{superseded_by}'.\n"
                        + "Note that you could enable `auto_supersede` to automatically\n"
                        + "load the superseding simulation.  Alternatively, you could\n"
                        + "pass `ignore_deprecation=True` or specify a version to load\n"
                        + "this waveform anyway."
                    )
                else:
                    raise ValueError(
                        f"Simulation '{sxs_id}' is superseded by '{superseded_by}'.\n"
                        + "Note that you could pass `ignore_deprecation=True` or\n"
                        + "specify a version to load this waveform anyway."
                    )
            if "deprecated" in metadata.get("keywords", []):
                raise ValueError(
                    f"Simulation '{sxs_id}' is deprecated but has no superseding simulation.\n"
                    + "Note that you could pass `ignore_deprecation=True` or specify a version\n"
                    + "to  to load this waveform anyway."
                )

    # Note the deprecation status in the kwargs, even if ignoring deprecation
    kwargs["deprecated"] = deprecated

    # We want to do this *after* deprecation checking, to avoid possibly unnecessary web requests
    files = get_file_info(metadata, sxs_id, download=kwargs.get("download", None))

    # If Lev is given as part of `location`, use it; otherwise, use the highest available
    lev_numbers = sorted({lev for f in files if (lev:=lev_number(f))})
    output_lev_number = input_lev_number or max(lev_numbers)
    location = f"{sxs_id_stem}{version}/Lev{output_lev_number}"

    # Finally, figure out which version of the simulation to load and dispatch
    version_number = float(version[1:])
    if 1 <= version_number < 2.0:
        return Simulation_v1(
            metadata, series, version, sxs_id_stem, sxs_id, url, files, lev_numbers, output_lev_number, location, *args, **kwargs
        )
    elif 2 <= version_number < 3.0:
        return Simulation_v2(
            metadata, series, version, sxs_id_stem, sxs_id, url, files, lev_numbers, output_lev_number, location, *args, **kwargs
        )
    else:
        raise ValueError(f"Version '{version}' not yet supported")


class SimulationBase:
    """Base class for Simulation objects

    Note that users almost certainly never need to call this function;
    see the `Simulation` function or `sxs.load` function instead.
    
    Attributes
    ----------
    metadata : Metadata
        Metadata object for the simulation
    series : pandas.Series
        The metadata, as extracted from the `simulations.dataframe`,
        meaning that it has columns consistent with other simulations,
        even when the underlying Metadata objects do not.  Note that
        `metadata` is an alias for this attribute, just based on the
        use of that name for `simulations`, but technically `pandas`
        distinguishes a single row like this as a `Series` object.
    version : str
        Version number of the simulation
    sxs_id_stem : str
        SXS ID without the version number or Lev
    sxs_id : str
        SXS ID with the version number
    location : str
        Location string for the simulation, including the SXS ID,
        version number, and Lev number.
    url : str
        URL for the DOI of the simulation
    files : dict
        Dictionary of file information for the simulation.  The keys
        are of the form "Lev5:Horizons.h5", and the values are
        dictionaries with keys "checksum", "size", and "link".
    lev_numbers : list
        List of available Lev numbers for the simulation.
    lev_number : int
        Chosen Lev number for the simulation.
    horizons : Horizons
        Horizons object for the simulation
    strain : Waveform
        Strain Waveform object for the simulation.  Note that `h` is
        an alias for this attribute, both of which are case
        insensitive: `Strain` and `H` are also acceptable.
    psi4 : Waveform
        Psi4 Waveform object for the simulation.  Note that this
        attribute is also case insensitive: `Psi4` is also acceptable.
    psi3 : Waveform
        Psi3 Waveform object for the simulation.  Note that this
        attribute is also case insensitive: `Psi3` is also acceptable.
        In versions 1 and 2, this attribute will raise an error
        because the data is not available.
    psi2 : Waveform
        Psi2 Waveform object for the simulation.  Note that this
        attribute is also case insensitive: `Psi2` is also acceptable.
        In versions 1 and 2, this attribute will raise an error
        because the data is not available.
    psi1 : Waveform
        Psi1 Waveform object for the simulation.  Note that this
        attribute is also case insensitive: `Psi1` is also acceptable.
        In versions 1 and 2, this attribute will raise an error
        because the data is not available.
    psi0 : Waveform
        Psi0 Waveform object for the simulation.  Note that this
        attribute is also case insensitive: `Psi0` is also acceptable.
        In versions 1 and 2, this attribute will raise an error
        because the data is not available.
    """
    def __init__(self,
        metadata, series, version, sxs_id_stem, sxs_id, url, files, lev_numbers, lev_number, location,
        *args, **kwargs
    ):
        self.metadata = metadata
        self.series = series
        self.version = version
        self.sxs_id_stem = sxs_id_stem
        self.sxs_id = sxs_id
        self.url = url
        self.files = files
        self.lev_numbers = lev_numbers
        self.lev_number = lev_number
        self.location = location
        self.deprecated = kwargs.get("deprecated", False)

    def __repr__(self):
        chi1 = self.metadata.reference_dimensionless_spin1
        chi2 = self.metadata.reference_dimensionless_spin2
        construction = f"""{type(self).__qualname__}("{self.location}")\n# """
        if self.deprecated:
            construction += "DEPRECATED "
        construction += f"{self.metadata.number_of_orbits:.3g}-orbit "
        construction += f"q={self.metadata.reference_mass_ratio:.3g} "
        construction += f"""chi1=[{", ".join(f"{c:.3g}" for c in chi1)}] """
        construction += f"""chi2=[{", ".join(f"{c:.3g}" for c in chi2)}] """
        construction += f"e={self.metadata.reference_eccentricity:.3g} simulation"
        return construction

    def __str__(self):
        return repr(self)
    
    @property
    def dataframe(self):
        return self.series

    @property
    def versions(self):
        return self.metadata.DOI_versions

    @property
    def lev(self):
        return f"Lev{self.lev_number}"

    @property
    def Lev(self):
        return self.lev

    def load_horizons(self):
        from .. import load
        sxs_id_path = Path(sxs_path_to_system_path(self.sxs_id))
        horizons_path = self.horizons_path
        horizons_location = self.files.get(horizons_path)["link"]
        horizons_truepath = sxs_id_path / sxs_path_to_system_path(horizons_path)
        return load(horizons_location, truepath=horizons_truepath)

    @property
    def horizons(self):
        if not hasattr(self, "_horizons"):
            self._horizons = self.load_horizons()
        return self._horizons
    Horizons = horizons

    @property
    def strain(self):
        if not hasattr(self, "_strain"):
            self._strain = self.load_waveform(*self.strain_path)
        return self._strain
    Strain = strain
    h = strain
    H = strain

    # I'm not entirely sure about the conjugations and factors of 2 in
    # shear and news in our conventions.  These will have to wait for
    # later.
    #
    # @property
    # def shear(self):
    #     if not hasattr(self, "_shear"):
    #         self._shear = self.strain.bar / 2
    #     return self._shear
    # sigma = shear
    # σ = shear
    # Shear = shear
    # Sigma = shear
    # Σ = shear
    #
    # @property
    # def news(self):
    #     if not hasattr(self, "_news"):
    #         self._news = self.strain.dot
    #     return self._news
    # News = news

    @property
    def psi4(self):
        if not hasattr(self, "_psi4"):
            self._psi4 = self.load_waveform(*self.psi4_path)
        return self._psi4
    Psi4 = psi4

    @property
    def psi3(self):
        raise AttributeError(f"Psi3 is not available for version {self.version} of the data")
    Psi3 = psi3

    @property
    def psi2(self):
        raise AttributeError(f"Psi2 is not available for version {self.version} of the data")
    Psi2 = psi2

    @property
    def psi1(self):
        raise AttributeError(f"Psi1 is not available for version {self.version} of the data")
    Psi1 = psi1

    @property
    def psi0(self):
        raise AttributeError(f"Psi0 is not available for version {self.version} of the data")
    Psi0 = psi0


class Simulation_v1(SimulationBase):
    """Simulation object for version 1 of the data format
    
    Note that users almost certainly never need to call this function;
    see the `Simulation` function or `sxs.load` function instead.  See
    also `SimulationBase` for the base class that this class inherits
    from.
    """
    # We have to deal with the fact that some early file paths on
    # Zenodo included the SXS ID as a prefix, while others did not.
    # This means that we have to check for both possibilities in
    # `load_horizons` and `load_waveform`.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extrapolation = kwargs.get("extrapolation", "N2")

    @property
    def horizons_path(self):
        return f"{self.lev}/Horizons.h5"
    
    def load_horizons(self):
        from .. import load
        sxs_id_path = Path(sxs_path_to_system_path(self.sxs_id))
        horizons_path = self.horizons_path
        if horizons_path in self.files:
            horizons_location = self.files.get(horizons_path)["link"]
        else:
            if (extended_horizons_path := f"{self.sxs_id_stem}/{horizons_path}") in self.files:
                horizons_location = self.files.get(extended_horizons_path)["link"]
            else:
                raise ValueError(f"File '{horizons_path}' not found in simulation files")
        horizons_truepath = sxs_id_path / sxs_path_to_system_path(horizons_path)
        return load(horizons_location, truepath=horizons_truepath)

    @property
    def strain_path(self):
        extrapolation = (
            f"Extrapolated_{self.extrapolation}.dir"
            if self.extrapolation != "Outer"
            else "OutermostExtraction.dir"
        )
        return (
            f"{self.lev}/rhOverM_Asymptotic_GeometricUnits_CoM.h5",
            extrapolation
        )

    @property
    def psi4_path(self):
        extrapolation = (
            f"Extrapolated_{self.extrapolation}.dir"
            if self.extrapolation != "Outer"
            else "OutermostExtraction.dir"
        )
        return (
            f"{self.lev}/rMPsi4_Asymptotic_GeometricUnits_CoM.h5",
            extrapolation
        )

    def load_waveform(self, file_name, group):
        from .. import load
        if file_name in self.files:
            location = self.files.get(file_name)["link"]
        else:
            if (extended_file_name := f"{self.sxs_id_stem}/{file_name}") in self.files:
                location = self.files.get(extended_file_name)["link"]
            else:
                raise ValueError(f"File '{file_name}' not found in simulation files")
        sxs_id_path = Path(sxs_path_to_system_path(self.sxs_id))
        truepath = sxs_id_path / sxs_path_to_system_path(file_name)
        w = load(location, truepath=truepath, extrapolation_order=group)
        w.metadata = self.metadata
        return w


class Simulation_v2(SimulationBase):
    """Simulation object for version 2 of the data format
        
    Note that users almost certainly never need to call this function;
    see the `Simulation` function or `sxs.load` function instead.  See
    also `SimulationBase` for the base class that this class inherits
    from.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extrapolation = kwargs.get("extrapolation", "N2")

    @property
    def horizons_path(self):
        return f"{self.lev}:Horizons.h5"

    @property
    def strain_path(self):
        return (
            f"{self.lev}:Strain_{self.extrapolation}",
            "/"
        )

    @property
    def psi4_path(self):
        extrapolation = (
            f"Extrapolated_{self.extrapolation}.dir"
            if self.extrapolation != "Outer"
            else "OutermostExtraction.dir"
        )
        return (
            f"{self.lev}:ExtraWaveforms",
            f"/rMPsi4_Asymptotic_GeometricUnits_CoM_Mem/{extrapolation}"
        )

    def load_waveform(self, file_name, group):
        from .. import load
        # Note that `name` should not have the file ending on input,
        # but we will strip it regardless with `.stem`.
        file_name = Path(file_name).stem
        sxs_id_path = Path(sxs_path_to_system_path(self.sxs_id))
        h5_path = f"{file_name}.h5"
        json_path = f"{file_name}.json"
        h5_location = self.files.get(h5_path)["link"]
        json_location = self.files.get(json_path)["link"]
        h5_truepath = sxs_id_path / sxs_path_to_system_path(h5_path)
        json_truepath = sxs_id_path / sxs_path_to_system_path(json_path)
        if not json_truepath.exists():
            if not read_config("download", True):
                raise ValueError(f"{json_truepath} not found and download is disabled")
            download_file(json_location, sxs_directory("cache") / json_truepath)
        return load(h5_location, truepath=h5_truepath, group=group, metadata=self.metadata)


def get_file_info(metadata, sxs_id, download=None):
    from .. import load_via_sxs_id
    if "files" in metadata:
        return metadata["files"]
    truepath = Path(sxs_path_to_system_path(sxs_id)) / "zenodo_metadata.json"
    record = load_via_sxs_id(sxs_id, "export/json", truepath=truepath, download=download)
    entries = record["files"]["entries"]
    return {
        str(filename): {
            "checksum": entry["checksum"],
            "size": entry["size"],
            "link": entry["links"]["content"],
        }
        for filename, entry in entries.items()
    }



import numpy as np
import quaternionic
from .. import WaveformModes

import juliacall


PostNewtonian = juliacall.newmodule("PN")
PostNewtonian.seval("using PostNewtonian")

from . import GWFrames

def pkg_update():
    """Update all installed Julia packages (but not julia itself)."""
    PostNewtonian.seval("import Pkg")
    return PostNewtonian.seval("Pkg.update()")


def PNWaveform(
        M1, M2, chi1, chi2, Omega_i, *,
        inertial=True,
        ell_min=2, ell_max=8, waveform_pn_order=None,
        **orbital_evolution_kwargs
):
    """Generate a PN waveform.

    The return value is an `sxs.WaveformModes` object with the
    following additional fields:

        - `M1` (array): The primary mass as a function of time.
        - `M2` (array): The secondary mass as a function of time.
        - `chi1` (array): The primary spin as a function of time.
        - `chi2` (array): The secondary spin as a function of time.
        - `frame` (array): The quaternionic frame as a function of
          time.
        - `v` (array): The orbital velocity as a function of time.
        - `orbital_phase` (array): The orbital phase as a function of
          time.

    This is a wrapper around the Julia functions
    `PostNewtonian.orbital_evolution` and
    `PostNewtonian.inertial_waveform`.  See [the Julia
    documentation](https://moble.github.io/PostNewtonian.jl/dev/internals/dynamics/#PostNewtonian.orbital_evolution)
    for details on the optional keyword arguments.

    Note that the full Julia interface is also accessible from this
    Python module.  See [the Julia
    docs](https://moble.github.io/PostNewtonian.jl/dev/interface/python/)
    for details.  Also, the GWFrames submodule of this module provides
    another interface.

    """
    # Integrate the orbital dynamics
    inspiral = PostNewtonian.orbital_evolution(
        M1, M2, chi1, chi2, Omega_i, **orbital_evolution_kwargs
    )
    values = PostNewtonian.stack(inspiral.u)

    # Compute the waveform in the inertial frame
    waveform_pn_order = waveform_pn_order or PostNewtonian.typemax(PostNewtonian.Int)
    coorbital_frame = quaternionic.array(values[8:12, :].to_numpy().T)
    if inertial:
        frame = np.array([quaternionic.one])
        frame_type = "inertial"
        h = PostNewtonian.inertial_waveform(
            inspiral, ell_min=ell_min, ell_max=ell_max, PNOrder=waveform_pn_order
        ).to_numpy().T
    else:
        frame = coorbital_frame
        frame_type = "coorbital"
        h = PostNewtonian.coorbital_waveform(
            inspiral, ell_min=ell_min, ell_max=ell_max, PNOrder=waveform_pn_order
        ).to_numpy().T

    w = WaveformModes(
        h,
        time=inspiral.t,
        modes_axis=1,
        ell_min=ell_min,
        ell_max=ell_max,
        spin_weight=-2,
        frame=frame,
        frame_type=frame_type,
        data_type="h",
        M1=values[0, :].to_numpy(),
        M2=values[1, :].to_numpy(),
        chi1=values[2:5, :].to_numpy().T,
        chi2=values[5:8, :].to_numpy().T,
        coorbital_frame=coorbital_frame,
        v=values[12, :].to_numpy(),
        orbital_phase=values[13, :].to_numpy(),
    )

    # Allow the extra fields to be accessed naturally
    w.M1 = w._metadata["M1"]
    w.M2 = w._metadata["M2"]
    w.chi1 = w._metadata["chi1"]
    w.chi2 = w._metadata["chi2"]
    w.coorbital_frame = w._metadata["coorbital_frame"]
    w.v = w._metadata["v"]
    w.orbital_phase = w._metadata["orbital_phase"]

    return w



"""Functions to load and save waveforms in SpECTRE CCE v1 format"""

from .. import WaveformModes


def save(*args, **kwargs):
    raise NotImplementedError("Saving waveforms in SpECTRE CCE format (v1) is not supported")


def load(file_name, **kwargs):
    """Load a waveform in SpECTRE CCE format (version 1)

    Parameters
    ----------
    file_name : str or Path
        Relative or absolute path to the input HDF5 file.  If this
        string contains but does not *end* with `'.h5'`, the remainder
        of the string is taken to be the group within the HDF5 file in
        which the data is stored.  Also note that a JSON file is
        expected in the same location, with `.h5` replaced by `.json`
        (and the corresponding data must be stored under the `group`
        key if relevant).

    Required keyword argument
    -------------------------
    group : str
        The group within the HDF5 file in which the data is stored.

    """
    import re
    from pathlib import Path
    import numpy as np
    import h5py
    import spherical

    # Make sure the file exists
    path = Path(file_name).expanduser().resolve()
    if not path.exists():
        raise FileNotFoundError(f"Could not find {path}")

    # Get the group name
    group = kwargs.pop("group", None)
    if group is None:
        raise ValueError("The 'group' keyword argument is required")

    # Determine the data type and spin weight from the group name
    if "news" in group.lower():
        data_type = "news"
        spin_weight = -2
    elif "psi0" in group.lower():
        data_type = "psi0"
        spin_weight = 2
    elif "psi1" in group.lower():
        data_type = "psi1"
        spin_weight = 1
    elif "psi2" in group.lower():
        data_type = "psi2"
        spin_weight = 0
    elif "psi3" in group.lower():
        data_type = "psi3"
        spin_weight = -1
    elif "psi4" in group.lower():
        data_type = "psi4"
        spin_weight = -2
    elif "strain" in group.lower():
        data_type = "h"
        spin_weight = -2
    else:
        raise ValueError(f"Unrecognized data type in group name '{group}'")

    m_is_scaled_out = True
    r_is_scaled_out = True

    with h5py.File(str(path), "r") as h5:
        if group is not None:
            h5 = h5[group]
        legend = list(h5.attrs["Legend"])
        if len(legend) != h5.shape[1]:
            raise ValueError(
                f"Number of columns in the data ({h5.shape[1]}) "
                f"does not match the number of entries in the legend ({len(legend)})"
            )
        regex = re.compile(r"(Real|Imag) Y_([0-9]+),([-0-9]+)")

        time_index = legend.index("time")
        time = h5[:, time_index]
        indices = np.argsort(time)
        time = time[indices]
        lm = np.array(
            [[m[2], m[3]] for l in legend if (m:=regex.match(l))],
            dtype=int
        )
        ell_min = min(lm[:, 0])
        ell_max = max(lm[:, 0])

        data = np.zeros((len(time), spherical.Ysize(ell_min, ell_max)), dtype=complex)
        for i, legend_entry in enumerate(legend):
            if (match:=regex.match(legend_entry)):
                ell, m = int(match[2]), int(match[3])
                if match[1] == "Real":
                    data[:, spherical.Yindex(ell, m, ell_min=ell_min)] += h5[:, i][indices]
                elif match[1] == "Imag":
                    data[:, spherical.Yindex(ell, m, ell_min=ell_min)] += 1j * h5[:, i][indices]
                else:
                    raise ValueError(f"Unrecognized legend entry '{legend_entry}'")

        return WaveformModes(
            data,
            time=time,
            time_axis=0,
            modes_axis=1,
            frame_type="inertial",
            data_type=data_type,
            m_is_scaled_out=m_is_scaled_out,
            r_is_scaled_out=r_is_scaled_out,
            ell_min=ell_min,
            ell_max=ell_max,
            spin_weight=spin_weight,
        )



"""Functions to convert waveform quantities to LVC format"""

import numpy as np
import h5py


def convert_modes(
    sxs_format_waveform,
    metadata,
    out_filename,
    modes,
    extrapolation_order="Extrapolated_N2",
    log=print,
    truncation_time=None,
    tolerance=5e-07,
    truncation_tol=None,
):
    """Computes amplitude and phase for an SXS-format waveform, computes
    ROM-spline for each mode, and writes to file.
    """
    from time import perf_counter
    from . import WaveformAmpPhase, Dataset
    from ... import load
    from ..dicts import KeyPassingDict
    
    extrap = str(extrapolation_order) + ".dir"

    if truncation_time is None:
        truncation_time = metadata["reference_time"] / (metadata["reference_mass1"] + metadata["reference_mass2"])

    h = load(sxs_format_waveform)
    if type(h) == KeyPassingDict:
        h = h[extrap]
    if not truncation_time is None:
        h = h[np.argmin(abs(h.t - truncation_time)) + 1 :]

    start_time = h.t[0]

    peak_time = h.max_norm_time()

    t = h.t - peak_time

    h_amp_phase = np.empty((h.data.shape[0], 2 * h.data.shape[1]))
    h_amp_phase[:, ::2] = h.abs
    h_amp_phase[:, 1::2] = h.arg_unwrapped

    if truncation_tol is True:  # Test for actual identity, not just equality
        truncation_tol = 5e-2 * tolerance
    elif truncation_tol is False:
        truncation_tol = None

    with h5py.File(out_filename, "w") as out_file:
        out_file.create_dataset("NRtimes", data=t)
        for i, mode in enumerate(modes):
            amp = h_amp_phase[:, 2 * i]
            phase = h_amp_phase[:, 2 * i + 1]

            phase_out = Dataset.from_data(t, phase, tolerance, error_scaling=amp, truncation_tol=truncation_tol)
            amp_out = Dataset.from_data(t, amp, tolerance, truncation_tol=truncation_tol)

            out_group_amp = out_file.create_group("amp_l{0[0]}_m{0[1]}".format(mode))
            amp_out.write(out_group_amp)
            out_group_phase = out_file.create_group("phase_l{0[0]}_m{0[1]}".format(mode))
            phase_out.write(out_group_phase)

    return start_time, peak_time, getattr(h, "version_hist", None)



import numpy as np
from scipy.optimize import root_scalar
from scipy.interpolate import CubicSpline, UnivariateSpline
import quaternionic
from ... import jit
from ..waveform_modes import WaveformModesDict

from numpy import exp
from numpy import pi as π
from quaternionic import z


@jit
def find_i1(f, i2):
    """Convenience function to find previous non-increasing index.
    
    Given index `i2` find the nearest preceding `i1` such that
    `f[i1-1] > f[i1]`; if there is no such `i1`, return 0.
    """
    for i in range(i2, 0, -1):
        if f[i] <= f[i-1]:
            return i
    return 0


def to_lvc_conventions(
        h, horizons, *,
        t_ref=None, f_ref=None,
        dt=None,
        f_low=None,
        ell_max=None,
        phi_ref=None, inclination=None,
        ell_max_epoch=None,
        **kwargs
):
    r"""Return SXS waveform in LVC conventions.

    This is the function underpinning `sxs.load_lvc`.  Essentially,
    that function is a fairly thin wrapper around this one; it just
    loads the `h` and `horizons` data and passes it to this function.
    For explanation of what this function does, the inputs, and
    outputs, see that function's docstring.

    """
    if t_ref is None and f_ref is None:
        raise ValueError("One of `t_ref` or `f_ref` must be specified")
    if t_ref is not None and f_ref is not None:
        raise ValueError("Only one of `t_ref` or `f_ref` may be specified")
    
    # Transform to corotating frame, if necessary
    if h.frame_type != "corotating":
        h = h.to_corotating_frame()
    
    # Find the spin data
    chi1 = horizons.A.chi_inertial
    chi2 = horizons.B.chi_inertial
    t_chi = chi1.t.copy()
    t_remnant = horizons.C.time.copy()
    mass_remnant = horizons.C.christodoulou_mass.ndarray

    # If requested, restrict `ell_max`
    if ell_max is not None:
        h = h[:, h.index(2,-2):h.index(ell_max, ell_max)+1]
        ell_max_epoch = min(ell_max_epoch or ell_max, ell_max)
    else:
        ell_max_epoch = ell_max_epoch or h.ell_max

    # Set `t` to equal 0 at the moment of maximum ell<=ell_max_epoch norm
    epoch_time = h[:, h.index(2,-2):h.index(ell_max_epoch, ell_max_epoch)+1].max_norm_time(interpolate=True)
    h.t -= epoch_time
    assert h.t[0] < 0.0 and h.t[-1] > 0.0, "Epoch time is not within waveform data"
    t_chi -= epoch_time
    t_remnant -= epoch_time

    # If `t_ref` is not specified, find `t_ref` from `f_ref`.
    # Otherwise, use `t_ref` to find `f_ref`.
    omega = h.frame.to_angular_velocity(h.t)  # Note that this is a vector-valued function of time
    f = np.linalg.norm(omega, axis=1) / (2 * np.pi)
    f = UnivariateSpline(h.t, f, s=sum((1e-4*f)**2))(h.t)  # Smooth the data a little
    i2 = h.max_norm_index()
    i1 = find_i1(f, i2)
    if f_low is not None and (f_low < f[i1] or f_low > f[i2]):
        raise ValueError(
            f"Requested frequency {f_low=} is outside this waveform's "
            f"frequency band ({f[i1]}, {f[i2]})"
        )
    if f_ref is not None and (f_ref < f[i1] or f_ref > f[i2]):
        raise ValueError(
            f"Requested frequency {f_ref=} is outside this waveform's "
            f"frequency band ({f[i1]}, {f[i2]})"
        )
    f_spline = CubicSpline(h.t[i1:i2], f[i1:i2])
    if f_ref is not None:
        t_ref = root_scalar(lambda t: f_spline(t)-f_ref, bracket=(h.t[i1], h.t[i2])).root
    else:
        f_ref = f_spline(t_ref)
    if f_low is not None:
        t_low = root_scalar(lambda t: f_spline(t)-f_low, bracket=(h.t[i1], h.t[i2])).root
    else:
        t_low = h.t[i1]
        f_low = f[i1]
        h = h[i1:, :]

    # Interpolate to find frame at `t_ref` and account for ϕ at that time
    h_ref = h.interpolate([t_ref])
    frame_ref = h_ref.frame[0]

    # Use omega at that time as a guess for the rough direction of the radiation axis,
    # and find the dominant principal axis of the <LL> matrix at that time
    omega_ref = omega[h.index_closest_to(t_ref)]
    rough_direction = omega_ref / np.linalg.norm(omega_ref)
    LL_axis = quaternionic.array.from_vector_part(
        h_ref.dominant_eigenvector_LL(rough_direction=rough_direction)[0]
    )

    # Now find the rotation *in the corotating frame* that aligns the
    # z axis with the radiation axis.
    R_ref = np.sqrt(-LL_axis * quaternionic.z)  # R_ref * z * R_ref.conj() ≈ LL_axis

    # If we rotate this waveform by `R_ref`, the LL_axis should be aligned
    # with the z-axis to machine precision.
    h_ref = h_ref.rotate(R_ref)

    # Now we augment the frame transformation to account for the
    # reference phase, as described in the docstring:
    #   * Im{h_{2,2}^sym} = 0  # There will be 4 values of phi that satisfy this
    #   * Re{h_{2,2}^sym} < 0  # This will reduce the number of solutions to 2
    #   * Im{h_{2,1}^sym} < 0  # An odd-m mode is required to break the degeneracy
    α = -np.angle(
        h_ref[0, h_ref.index(2,2)]
        + h_ref[0, h_ref.index(2,-2)].conjugate()
    ) / 2
    h_ref2 = h_ref.rotate(exp(z * α/2))
    β = (
        h_ref2[0, h_ref2.index(2,2)]
        + h_ref2[0, h_ref2.index(2,-2)].conjugate()
    )
    if β.real > 0:
        α += π / 2
    h_ref3 = h_ref.rotate(exp(z * α/2))
    γ = (
        h_ref3[0, h_ref3.index(2,1)]
        + h_ref3[0, h_ref3.index(2,-1)].conjugate()
    )
    if γ.imag > 0:
        α += π
    R_ref = R_ref * exp(z * α/2)
    
    # We want to re-interpret the corotating waveform so that the
    # waveform *in this frame* remains the same, but appears to
    # inertial observers to be moving with an angular velocity (as a
    # function of time) that is just rotated by a constant.  This
    # means that we need to multiply `h.frame` *on the left* by a
    # constant.  And we want the resulting frame as a function of time
    # to be `R_ref.conj()` at time `t_ref`.
    R_adjustment = R_ref.conj() * frame_ref.conj()
    frame_quat = R_adjustment * h.frame
    h._metadata["frame"] = frame_quat

    # Transform to inertial-frame waveform, spins, and omega to `t_ref` frame
    h = h.to_inertial_frame()
    omega = np.array([R_adjustment.rotate(v) for v in omega])
    chi1 = np.array([R_adjustment.rotate(v) for v in chi1])
    chi2 = np.array([R_adjustment.rotate(v) for v in chi2])
    chi_remnant = np.array([R_adjustment.rotate(v) for v in horizons.C.chi_inertial])

    # If requested, interpolate to time step `dt`, ensuring that `t=0` is preserved
    if dt is not None:
        t = np.concatenate((
            np.arange(0.0, h.t[0], -dt)[::-1],
            np.arange(dt, h.t[-1], dt)
        ))
        h = h.interpolate(t)

    # If either `phi_ref` or `inclination` is not None, set default for the other
    if phi_ref is not None:
        inclination = inclination or 0.0
    if inclination is not None:
        phi_ref = phi_ref or 0.0

    chi1_ref = chi1.interpolate([t_ref+epoch_time]).ndarray[0]
    chi2_ref = chi2.interpolate([t_ref+epoch_time]).ndarray[0]

    dynamics_dict = {
        "t_ref": t_ref,
        "f_ref": f_ref,
        "t_low": t_low,
        "f_low": f_low,
        "chi1_ref": chi1_ref,
        "chi2_ref": chi2_ref,
        "frame_quat": frame_quat,
        "frame_omega": omega,
        "times_spins": t_chi,
        "chi1": chi1,
        "chi2": chi2,
        "times_remnant": t_remnant,
        "chi_remnant": chi_remnant,
        "mass_remnant": mass_remnant,
    }

    # If `phi_ref` and `inclination` are not None, return polarizations
    if phi_ref is not None:
        hp, hc = h.evaluate(inclination, π/2 - phi_ref).ndarray.view((float, 2)).T
        return t, hp, hc, dynamics_dict
    else:
        # Could do `dict(WaveformModesDict(h))` to convert to a plain dict
        hlm_dict = WaveformModesDict(h)
        return t, hlm_dict, dynamics_dict



import pytest
import sxs


@pytest.mark.parametrize("loader", [sxs.Simulation, sxs.load])
def test_superseded_by_single(loader):
    simulation = "SXS:BBH:0001"
    # NOTE: These tests as written will only work so long as only one
    # simulation is listed in the given simulation's `superseded_by`
    # field.  If more are added, this test will need to be changed to
    # use a simulation that still has just one.
    with pytest.raises(ValueError):
        loader(f"{simulation}")
    with pytest.warns(UserWarning):
        s = loader(f"{simulation}v2.0")
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}", ignore_deprecation=True)
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}v2.0", ignore_deprecation=True)
    assert s.sxs_id_stem == simulation
    with pytest.warns(UserWarning):
        s = loader(f"{simulation}", auto_supersede=True)
    assert s.sxs_id_stem != simulation
    with pytest.warns(UserWarning):
        s = loader(f"{simulation}v2.0", auto_supersede=True)
    assert s.sxs_id_stem != simulation
    s = loader(f"{simulation}", ignore_deprecation=True, auto_supersede=True)
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}v2.0", ignore_deprecation=True, auto_supersede=True)
    assert s.sxs_id_stem == simulation


@pytest.mark.parametrize("loader", [sxs.Simulation, sxs.load])
def test_superseded_by_multiple(loader):
    simulation = "SXS:BBH:0030"
    # NOTE: These tests assume that multiple simulations are listed in
    # the given simulation's `superseded_by` field.  If that changes
    # for some reason, this test will need to be changed to use a
    # simulation that still has multiple simulations.
    with pytest.raises(ValueError):
        loader(f"{simulation}")
    with pytest.warns(UserWarning):
        s = loader(f"{simulation}v2.0")
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}", ignore_deprecation=True)
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}v2.0", ignore_deprecation=True)
    assert s.sxs_id_stem == simulation
    with pytest.raises(ValueError):
        loader(f"{simulation}", auto_supersede=True)
    with pytest.raises(ValueError):
        loader(f"{simulation}v2.0", auto_supersede=True)
    s = loader(f"{simulation}", ignore_deprecation=True, auto_supersede=True)
    assert s.sxs_id_stem == simulation
    s = loader(f"{simulation}v2.0", ignore_deprecation=True, auto_supersede=True)
    assert s.sxs_id_stem == simulation



"""Functions to facilitate generic handling of SXS-format data files"""

import contextlib
from . import waveforms, doi_url


class JSONHandler:
    """Utility for loading and saving JSON files"""
    @classmethod
    def load(cls, file, **kwargs):
        import json
        if hasattr(file, "read"):
            return json.load(file, **kwargs)
        with open(file, "r") as f:
            return json.load(f, **kwargs)
    @classmethod
    def save(cls, obj, file, **kwargs):
        import json
        if hasattr(file, "write"):
            return json.dump(obj, file, **kwargs)
        with open(file, "w") as f:
            return json.dump(obj, f, **kwargs)


def sxs_handler(format_string):
    """Find an object to load from or save to files in the given format.

    Parameters
    ----------
    format_string : str

    Returns
    -------
    handler : object
        This object will have (at least) two attributes: `load` and `save`, which
        can be called as `handler.load(file, **kwargs)` and `handler.save(obj,
        file, **kwargs)`.

    See Also
    --------
    sxs_loader : Returns the function that will load a given file
    sxs.utilities.file_format : Returns just the string found in the file

    """
    import itertools
    import re
    from . import catalog, metadata, horizons, waveforms

    if not format_string:
        raise ValueError("Empty string cannot be associated with a handler")
    elif format_string.lower().startswith("catalog"):
        format_string = re.sub(r"^catalog\.?", "", format_string, count=1, flags=re.IGNORECASE)
        return catalog.formats.get(format_string, catalog.formats[None])
    elif format_string.lower().startswith("metadata"):
        format_string = re.sub(r"^metadata\.?", "", format_string, count=1, flags=re.IGNORECASE)
        return metadata.formats.get(format_string, metadata.formats[None])
    elif format_string.lower().startswith("horizons"):
        format_string = re.sub(r"^horizons\.?", "", format_string, count=1, flags=re.IGNORECASE)
        return horizons.formats.get(format_string, horizons.formats[None])
    elif format_string.lower().startswith("waveforms"):
        format_string = re.sub(r"^waveforms\.?", "", format_string, count=1, flags=re.IGNORECASE)
        return waveforms.formats.get(format_string, waveforms.formats[None])
    elif format_string.lower() == "json":
        return JSONHandler
    else:
        format_list = [
            catalog.formats,
            metadata.formats,
            horizons.formats,
            waveforms.formats,
        ]
        format_cycler = itertools.cycle(format_list)
        for _ in range(len(format_list)):
            format_dict = next(format_cycler)
            if format_string in format_dict:
                if any(format_string in next(format_cycler) for _ in range(len(format_list)-1)):
                    raise ValueError(f"Format string '{format_string}' found in multiple sxs format groups")
                return format_dict[format_string]
    raise ValueError(f"Format '{format_string}' is unknown to the `sxs` package; maybe you need to update `sxs`")


def sxs_loader(file, group=None):
    """Find the function that will load the given file

    If a format is specified, we assume that it is a top-level attribute or member
    named "sxs_format" or just "format".  If neither exists, return None; the
    calling function should check for this possibility.

    Parameters
    ----------
    file : file-like object, string, or pathlib.Path
    group : str, optional
        If the file is an HDF5 or JSON file, this is the group to inspect.

    Returns
    -------
    load : callable
        This function can be called as `load(file, **kwargs)`.

    See Also
    --------
    sxs_handler : Returns an object to load from and save to a given format
    sxs.utilities.file_format : Returns just the string found in the file

    Notes
    -----
    Note that this function has a very different signature than the related
    functions `sxs_handler`, which takes only the format string, not an actual
    file.

    """
    import re
    import pathlib
    from .utilities import file_format
    format_string = file_format(file, group)
    if format_string is None:
        file_string = str(pathlib.Path(file).name).lower()
        if "catalog" in file_string:
            format_string = "catalog"
        elif file_string.startswith("metadata"):
            format_string = "metadata"
        elif "horizons" in file_string:
            format_string = "horizons"
        elif re.match("(rh_|rhoverm_|rpsi4_|rmpsi4_)", file_string):
            format_string = "waveforms"
        elif file_string.endswith("json"):
            format_string = "json"
        else:
            raise ValueError(f"File '{file}' contains no recognized format information")
    handler = sxs_handler(format_string)
    # noinspection PyUnresolvedReferences
    return handler.load


def _safe_resolve_exists(path):
    """Evaluate `path.resolve().exists()`  without throwing exception
    
    This is just here to work around a bug that turned up in Windows
    on python 3.8.  It's not clear if it turns up in other versions.
    """
    try:
        return path.resolve().exists()
    except:
        return False

def load(location, download=None, cache=None, progress=None, truepath=None, **kwargs):
    """Load an SXS-format dataset, optionally downloading and caching

    The dataset can be the full catalog of all SXS simulations, or
    metadata, horizon data, or a waveform from an individual
    simulation.

    Parameters
    ----------
    location : {str, pathlib.Path}
        A local file path, URL, SXS path, or SXS path pattern.  See
        Notes below.
    download : {None, bool}, optional
        If this is True and the data is recognized as starting with an
        SXS ID but cannot be found in the cache, the data will be
        downloaded automatically.  If this is None (the default) and
        an SXS configuration file is found with a `download` key, that
        value will be used.  If this is False, any configuration will
        be ignored, and no files will be downloaded.  Note that if
        this is True but `cache` is None, `cache` will automatically
        be switched to True.
    cache : {None, bool}, optional
        The cache directory is determined by `sxs.sxs_directory`, and
        any downloads will be stored in that directory.  If this is
        None (the default) and `download` is True it will be set to
        True.  If this is False, any configuration will be ignored and
        any files will be downloaded to a temporary directory that
        will be deleted when python exits.
    progress : {None, bool}, optional
        If True, full file names will be shown and, if a nonzero
        Content-Length header is returned, a progress bar will be
        shown during any downloads.  Default is None, which just reads
        the configuration value with `read_config("download_progress",
        True)`, defaulting to True.
    truepath : {None, str}, optional
        If the file is downloaded, this allows the output path to be
        overridden, rather than selected automatically.  The output
        path will be stored in `truepath` relative to the cache
        directory.

    Keyword Parameters
    ------------------
    All remaining parameters are passed to the `load` function
    responsible for the requested data.

    See Also
    --------
    sxs.sxs_directory : Locate configuration and cache files
    sxs.write_config : Set defaults for `download` and `cache`
        parameters

    Notes
    -----
    This function can load data in various ways.

      1) Given an absolute or relative path to a local file, it just
         loads the data directly.

      2) If `truepath` is set, and points to a file that exists —
         whether absolute, relative to the current working directory,
         or relative to the cache directory — that file will be
         loaded.

      3) If `location` is a valid URL including the scheme (https://,
         or http://), it will be downloaded regardless of the
         `download` parameter and optionally cached.

      4) Given an SXS simulation specification — like "SXS:BBH:1234",
         "SXS:BBH:1234v2.0", "SXS:BBH:1234/Lev5", or
         "SXS:BBH:1234v2.0/Lev5" — the simulation is loaded as an
         `sxs.Simulation` object.

      5) Given an SXS path — like
         "SXS:BBH:1234/Lev5/h_Extrapolated_N2.h5" — the file is
         located in the catalog for details.  This function then looks
         in the local cache directory and loads it if present.

      6) If the SXS path is not found in the cache directory and
         `download` is set to `True` (when this function is called, or
         in the sxs config file) this function attempts to download
         the data.  Note that `download` must be explicitly set in
         this case, or a ValueError will be raised.

    If the file is downloaded, it will be stored in the cache
    according to the `location`, unless `truepath` is set as noted
    above, in which case it is stored there.  Note that downloading is
    switched off by default, but if it is switched on (set to True),
    the cache is also switched on by default.

    """
    import pathlib
    import urllib.request
    from . import Simulations, Simulation, read_config, sxs_directory, Catalog
    from .utilities import url, download_file, sxs_path_to_system_path, sxs_id_version_lev_exact_re

    # Note: `download` and/or `cache` may still be `None` after this
    if download is None:
        download = read_config("download", True)
    if cache is None:
        cache = read_config("cache")
    if progress is None:
        progress = read_config("download_progress", True)

    # We set the cache path to be persistent if `cache` is `True` or `None`.  Thus,
    # we test for whether or not `cache` literally *is* `False`, rather than just
    # if it casts to `False`.
    cache_path = sxs_directory("cache", persistent=(cache is not False))

    path = pathlib.Path(sxs_path_to_system_path(location)).expanduser()  # .resolve()
    h5_path = path.with_suffix('.h5')
    json_path = path.with_suffix('.json')

    if not path.exists():
        if truepath and (testpath := pathlib.Path(truepath).expanduser()).exists():
            path = testpath

        elif truepath and (testpath := cache_path / truepath).exists():
            path = testpath

        elif _safe_resolve_exists(h5_path):
            path = h5_path

        elif _safe_resolve_exists(json_path):
            path = json_path

        elif "scheme" in url.parse(location):
            m = url.parse(location)
            truepath = truepath or urllib.request.url2pathname(f"{m['host']}/{m['port']}/{m['resource']}")
            path = cache_path / truepath
            if not path.resolve().exists():
                if download is False:  # Again, we want literal False, not casting to False
                    raise ValueError(f"File '{truepath}' not found in cache, but downloading turned off")
                download_file(location, path, progress=progress)

        elif location == "catalog":
            return Catalog.load(download=download)

        elif location == "simulations":
            return Simulations.load(download=download)
        
        elif sxs_id_version_lev_exact_re.match(location):
            return Simulation(location, download=download, cache=cache, progress=progress, **kwargs)

        else:
            # Try to find an appropriate SXS file in the catalog
            catalog = Catalog.load(download=download)
            selections = catalog.select_files(location)
            if not selections:
                raise ValueError(f"Nothing found matching '{location}'")
            if progress:
                print("Found the following files to load from the SXS catalog:")
                print("    " + "\n    ".join(selections))
            paths = []
            for sxs_path, file_info in selections.items():
                truepath = truepath or sxs_path_to_system_path(file_info.get("truepath", sxs_path))
                path = cache_path / truepath
                if not path.resolve().exists():
                    download_url = file_info["download"]
                    download_file(download_url, path, progress=progress)
                paths.append(path)
            loaded = [load(path, download=False, progress=progress, **kwargs) for path in paths]
            if len(loaded) == 1:
                return loaded[0]
            else:
                return loaded

    loader = sxs_loader(path, kwargs.get("group", None))

    return loader(path, **kwargs)


def load_via_sxs_id(sxsid, location, *, download=None, cache=None, progress=None, truepath=None, **kwargs):
    """Load a path via a (possibly versioned) SXS ID
    
    Given some SXS ID like "SXS:BBH:1234" or a versioned ID like
    "SXS:BBH:1234v2.0", we may wish to first resolve the DOI to a
    specific Zenodo record, and then load a specific path under that
    record — for example, we may want to export the Zenodo record as
    JSON by appending "export/json" to the Zenodo URL, which we
    *could* get as

        load("https://zenodo.org/records/13152488/export/json")

    because that's the record "SXS:BBH:1234v2.0" resolves to.
    However, we don't want to keep track of the Zenodo URL, so we
    just use this function instead, as

        load_via_sxs_id("SXS:BBH:1234v2.0", "export/json")
    
    """
    from pathlib import Path
    import requests
    from .utilities import sxs_path_to_system_path
    url = f"{doi_url}{sxsid}"
    response = requests.head(url, allow_redirects=True)
    if response.status_code != 200:
        raise ValueError(f"Could not load via DOI {url=}")
    final_url = f"{response.url}/{location}"
    truepath = truepath or Path(sxs_path_to_system_path(sxsid)) / location
    return load(final_url, download, cache, progress, truepath, **kwargs)


@contextlib.contextmanager
def loadcontext(*args, **kwargs):
    """Context manager for backwards compatibility

    This context manager takes precisely the same arguments as `sxs.load` and
    yields precisely the same results; essentially it is a trivial wrapper around
    the `load` function.  The benefit of this approach is that it can be used in
    precisely the same way as `h5py.File` would have been used previously.  For
    example, the old approach would be to open an HDF5 file like this:

        with h5py.File("Horizons.h5", "r") as horizons:
            areal_mass = horizons["AhA.dir/ArealMass.dat"]

    With this function, the same effect can be achieved as

        with sxs.loadcontext("Horizons.h5") as horizons:
            areal_mass = horizons["AhA.dir/ArealMass.dat"]

    Each of the datasets found in Horizons.h5 as well as the old NRAR-style files
    will be available through this interface, even when using newer files in
    different formats.  Thus, only one line of code would need to change to use the
    new interface.

    However, be aware that this may not an be efficient use of memory, and is
    almost certainly slower than the newer interfaces.  Wherever possible, you
    should update your code to use newer interfaces.  Failing to do so will leave
    you open to ridicule from your peers and loved ones.

    See Also
    --------
    load

    """
    yield load(*args, **kwargs)


def load_lvc(
        sxs_id, *,
        t_ref=None, f_ref=None,
        dt=None,
        f_low=None,
        ell_max=None,
        phi_ref=None, inclination=None,
        ell_max_epoch=None,
        **kwargs
):
    r"""Load an SXS waveform in LVC convention.

    Returns an SXS waveform (modes or polarizations) and dynamics
    (including angular velocities, frame quaternions, and spins) in
    the inertial frame that coincides with the waveform-defined frame
    defined at a reference time `t_ref` or reference frequency
    `f_ref`.

    Parameters
    ==========
    sxs_id : str
        The SXS ID of the simulation to use — e.g., "SXS:BBH:1234".
    t_ref : float, optional
        The reference time at which the waveform frame is specified.
        This is measured in units of M, and defined relative to the
        epoch time (see below).  Either `t_ref` or `f_ref` must be
        specified.  If `t_ref` is given, it is used to compute
        `f_ref`.
    f_ref : float, optional
        The reference frequency, in units of cycles/M, at which the
        waveform frame is specified.  Either `t_ref` or `f_ref` must
        be specified.  If `f_ref` is given, it is used to compute
        `t_ref`.
    dt : float, optional
        The time step, in units of M, to which to interpolate the
        waveform.
    f_low : float, optional
        The lower frequency bound, in units of cycles/M, for the
        waveform.
    ell_max : int, optional
        The maximum ell to include in the waveform.
    phi_ref : float, optional
        The binary's phase in the coprecessing frame, measured at
        `t_ref`. Should be between 0 and $2\pi$.
    inclination : float, optional
        Angle between the binary's angular momentum and the line of
        sight of the observer, measured at `t_ref`.  Should be between
        0 and $\pi$.
    ell_max_epoch : int, optional
        The maximum ell to include in the epoch time calculation,
        which sets t=0 at the maximum of the L^2 norm, calculated by
        including all modes up to and including this ell value.

    Returns
    =======
    times : float array
        Uniformly spaced 1D array of times, in units of M, at which
        the waveform and dynamics quantities are returned.  Aligned
        such that peak of waveform modes with ell=2 is at t=0.
    hlm_dict : dict [optional]
        Dictionary of waveform modes in the inertial frame that
        coincides with the waveform-defined coprecessing frame at
        `f_ref`.  Each mode in the dictionary is a 1D array of
        complex-valued floats with values corresponding to each time
        node.  Keys:[(ell,m)] for all ell<=ell_max and -ell<=m<=+ell.
        This is returned only if the input values of `phi_ref` and
        `inclination` are both None.
    hp, hc : float arrays [optional]
        1D-arrays of real-valued GW polarizations evaluated in the
        frame of the observer at each time node.  Polarizations are
        computed using all modes up to ell=ell_max, with one value at
        each time node.  These are returned only if either of the
        input values of `phi_ref` and `inclination` is not None.
    dynamics_dict : dict
        Dictionary of real-valued arrays of dynamics quantities:
            * "t_ref": The reference time at which the waveform frame
              is specified.
            * "f_ref": The waveform's frequency at `t_ref`.
            * "t_low": The earliest time in the waveform.
            * "f_low": The waveform's frequency at `t_low`.
            * "chi1_ref": Cartesian spin components for the more
              massive object, evaluated at `t_ref` in the
              waveform-defined inertial frame.
            * "chi2_ref": Cartesian spin components for the less
              massive object, evaluated at `t_ref` in the
              waveform-defined inertial frame.
            * "frame_quat": Quaternions describing the instantaneous
              transformation from the inertial frame to the corotating
              frame at each time node as described in arXiv:1905.09300
              and using conventions in Appendix B of arXiv:1110.2965.
              Array of shape (len(times),4) with four quaternion
              components at each time node.  The first element at each
              time node is the scalar part.
            * "frame_omega": Angular velocity vector of the corotating
              frame at each time node.  Array of shape (len(times),3)
              with three components at each time node.
            * "times_spins": The times, in units of M, at which `chi`
              and `chi2` are returned.  Note that these are coordinate
              times deep within the dynamic region of the simulation,
              and so cannot be precisely related to the times in the
              asymptotic waveform.
            * "chi1": Cartesian spin components for the more massive
              object, evaluated in the waveform-defined inertial
              frame.  Array of shape (len(times_spins),3) which
              contains the Cartesian spin components
              \{chi_{1x},chi_{1y},chi_{1z}\} at each time node.
            * "chi2": Cartesian spin components for the less massive
              object, evaluated in the waveform-defined inertial
              frame.  Array of shape (len(times_spins),3) which
              contains the Cartesian spin components
              \{chi_{2x},chi_{2y},chi_{2z}\} at each time node.
            * "times_remnant": The times, in units of M, at which
              `chi_remnant` and `mass_remnant` are returned.  Note
              that these are coordinate times deep within the dynamic
              region of the simulation, and so cannot be precisely
              related to the times in the asymptotic waveform.
            * "chi_remnant": Cartesian spin components for the remnant
              black hole, evaluated in the waveform-defined inertial
              frame.  Array of shape (len(times_remnant),3) which
              contains the Cartesian spin components
              \{chi_{rx},chi_{ry},chi_{rz}\}.
            * "mass_remnant": The Christodoulou mass of the remnant
              black hole as a function of time.  Array of shape
              (len(times_remnant),).

    See also
    ========
    sxs.load : General-purpose function to load SXS data in native
        format
    sxs.waveforms.to_lvc_conventions : Inner function that does all
        the work for this function

    Conventions
    ===========
    We assume geometric units for time (units of M) and frequency
    (units of cycles/M), with total mass M equal to 1.
    
    Epoch time is defined by the peak of the $L^2$ norm of the modes:

        $$t_e = \argmax_t \sum_\ell \sum_m |h_{\ell,m}(t)|^2,$$

    where the sum over $\ell$ ranges from 2 to `ell_max_epoch`.  All
    time axes are then adjusted so that $t_e = 0$.

    Frequencies are measured from the waveform, rather than orbital
    trajectories, in terms of the angular-velocity vector given by
    equation (7) of arXiv:1302.2919.  The frequency is defined as the
    magnitude of this vector divided by $2\pi$.

    Waveforms, spins, dynamics, times, inclination angle and GW
    reference phase are all defined in the inertial frame that
    coincides with the "waveform-defined frame" at the reference time.
    This frame is chosen so that the $z$ axis is aligned with the
    dominant principal axis of the matrix given by equation (2a) of
    arXiv:1109.5224, except that the strain is used in place of
    $\psi_4$.  That axis is only determined up to a sign, so we choose
    the positive $z$ axis to be more parallel than antiparallel to the
    angular velocity.  Rotation about the $z$ axis is chosen to
    approximate the condition that the more massive black hole is
    located on the positive $x$ axis at the reference time, but can be
    written solely in terms of the waveform modes:

        * $\Im{h_{2,2} + \bar{h}_{2,-2}} = 0$
        * $\Re{h_{2,2} + \bar{h}_{2,-2}} < 0$
        * $\Im{h_{2,1} + \bar{h}_{2,-1}} < 0$

    The first two conditions are necessary for cases of symmetric
    systems in which $h_{2,\pm 1}=0$; the last condition breaks the
    degeneracy of the first two under rotation about the $z$ axis by
    $\pi$.  For configurations that are symmetric under that rotation,
    $h_{2,1}$ will be zero, so this condition will be impossible to
    apply, but the symmetry of the system will mean that there is no
    difference in the result.

    Quaternion conventions are described in Appendix B of
    arXiv:1110.2965.  In particular, we use the convention that

        Q = q_0 + vec(q) = (q_0, q_1, q_2, q_3)

    where q_0 is the scalar part.

    """
    lev = kwargs.pop("lev", "Lev")  # If number is unspecified, load chooses highest
    waveform_name = kwargs.pop("waveform_name", "Strain_N2.h5")
    h = load(f"{sxs_id}/{lev}/{waveform_name}", transform_to_inertial=False)
    horizons = load(f"{sxs_id}/{lev}/Horizons.h5")
    return waveforms.to_lvc_conventions(
        h, horizons,
        t_ref=t_ref, f_ref=f_ref,
        dt=dt,
        f_low=f_low,
        ell_max=ell_max,
        phi_ref=phi_ref, inclination=inclination,
        ell_max_epoch=ell_max_epoch,
        **kwargs
    )



"""Functions to load and save waveforms in RPXMB format"""

import sys
import warnings
import tempfile
import contextlib
import pathlib
import bz2
import json
import numpy as np
import scipy
import h5py
import quaternionic
import spherical

from .. import WaveformModes
from . import rotating_paired_diff_multishuffle_bzip2 as rpdmb
from ... import __version__
from ...utilities import default_shuffle_widths, md5checksum, xor, multishuffle, version_info


sxs_formats = ["rotating_paired_xor_multishuffle_bzip2", "rpxmb", "rpxm", "RPXMB", "RPXM"]


def save(w, file_name=None, file_write_mode="w", L2norm_fractional_tolerance=1e-10, log_frame=None, shuffle_widths=default_shuffle_widths):
    """Save a waveform in RPXMB format

    This function is primarily for backwards compatibility.  The `RPDMB` format is
    now preferred.  In fact, this function just calls `rpdmb.save`, with two
    non-default parameters, to look for `RPXMB` format strings, and to use `xor` in
    place of `diff`.  See that function's docstring for more details.

    Also note that the default `shuffle_widths` parameter has changed.  The
    previous default is now available as
    `sxs.utilities.default_shuffle_widths_old`.  The `load` function will
    automatically read the shuffle widths used from the saved file, so you do not
    need to worry about this unless you are comparing checksums or something like
    that.

    """
    return rpdmb.save(
        w, file_name=file_name, file_write_mode=file_write_mode,
        L2norm_fractional_tolerance=L2norm_fractional_tolerance, log_frame=log_frame,
        shuffle_widths=shuffle_widths, diff=xor,
        formats=sxs_formats,
    )


def load(file_name, ignore_validation=None, check_md5=True, transform_to_inertial=True, **kwargs):
    """Load a waveform in RPXMB format

    This function is primarily for backwards compatibility.  The `RPDMB` format is
    now preferred.  In fact, this function just calls `rpdmb.load`, with two
    non-default parameters, to look for `RPXMB` format strings, and to use `xor` in
    place of `diff`.  See that function's docstring for more details.

    """
    return rpdmb.load(
        file_name, ignore_validation=ignore_validation, check_md5=check_md5,
        transform_to_inertial=transform_to_inertial, diff=xor, formats=sxs_formats,
        **kwargs
    )
